from .VisBeatImports import *
from .Video import *
from .AFileManager import AFileManager

try:
    import youtube_dl
except ImportError:
    AWARN('You need to install youtube-dl to use parts of VideoSource class that pull from YouTube. Try running:\npip install youtube-dl')



def safe_file_name(input_string):
    return ''.join([i if ord(i) < 128 else '_' for i in input_string]);

class VideoSource(AFileManager):
    """Video (class): A video, and a bunch of convenience functions to go with it.
        Attributes:
            name: The name of the video
            sampling_rate: The framerate of the video
            audio: an Audio object, the audio for the video
    """

    VIDEO_TEMP_DIR = './'

    def getJSONName(self):
        return self.AOBJECT_TYPE()+".json";

    def AOBJECT_TYPE(self):
        return 'VideoSource';

    def __init__(self, path, name=None, source_location=None, VideoClass = None, **kwargs):
        """
        :param path: either the path to a json, or the path to a directory containing 'VideoSource.json',
            or the path to a directory where said json should be created.
        :param name: name for video
        :param source_location: can be a path to a video file, or a youtube source from which to pull a video file
        :param VideoClass: visbeat.Video by default, but can be changed if you are using a custom subclass.
               Must be a subclass of visbeat.Video, and must be constructable with VideoClass(path)
        """
        AFileManager.__init__(self, path=path);

        if(VideoClass is not None):
            self.VideoClass = VideoClass;

        self.setSource(source_location=source_location, assert_valid=None, **kwargs);

        # if (self.name is None):
            # self.name = os.path.splitext(os.path.basename(path))[0]


    @staticmethod
    def NewVideoSource(destination, name, source_location=None, VideoClass = None, **kwargs):
        vsdir = os.path.join(destination, name+os.sep);
        make_sure_dir_exists(vsdir);
        print(("Video source at {}".format(vsdir)));
        return VideoSource(path=vsdir, name=name, source_location=source_location, **kwargs);

    def initializeBlank(self):
        AFileManager.initializeBlank(self);
        self.source_location = None;
        self.name = None;
        self.video_file_name = None;
        self.title_safe = None;
        self.youtube_info_dict = None;
        self.versions_info = {};

        #
        self.VideoClass = Video;

    def toDictionary(self):
        d = AFileManager.toDictionary(self);
        d['source_location']=self.source_location;
        if(self.name):
            d['name']=self.name;
        if(self.video_file_name):
            d['video_file_name']=self.video_file_name;
        if(self.title_safe):
            d['title_safe']=self.title_safe;
        if(self.youtube_info_dict):
            d['youtube_info_dict']=self.youtube_info_dict;
        if(self.versions_info):
            d['versions_info']=self.versions_info;
        return d;


    def initFromDictionary(self, d):
        AFileManager.initFromDictionary(self, d);
        self.source_location = d['source_location'];
        self.name = d.get('name');
        self.video_file_name = d.get('video_file_name');
        self.title_safe = d.get('title_safe');
        self.youtube_info_dict = d.get('youtube_info_dict');
        self.versions_info=d.get('versions_info');
        if(self.versions_info is None):
            self.versions_info = {};
        #do class specific inits with d;


    def initWithPath(self, path=None, clear_temp=None):
        AFileManager.initWithPath(self, path=path, clear_temp=clear_temp);
        self.setDir('versions', pathstring(self.getDirectoryPath() + os.sep + "Versions" + os.sep));
        self.setDir('warps', pathstring(self.getDir('versions')+ os.sep + "Warps" + os.sep));
        # self.setDir('midi', pathstring(self.getDirectoryPath() + os.sep + "MIDI" + os.sep));
        # self.setDir('music', pathstring(self.getDirectoryPath() + os.sep + "Music" + os.sep));
        self.setFeaturesDir();

    def setFeaturesDir(self, features_dir=None):
        if (features_dir is None):
            self.setDir('features', pathstring(self.getDir('data') + os.sep + "Features" + os.sep))
        else:
            self.setDir('features', features_dir);



    def getName(self):
        """
        Gets name if set. If not set, returns file name without extension.
        :return:
        """
        if(self.name is not None):
            return self.name;
        elif(self.video_file_name is not None):
            return os.path.splitext(self.video_file_name)[0];
        else:
            return None;


###################


    @staticmethod
    def _versionLabelString(version_label=None):
        if (version_label and version_label != 'Full'):
            return str(version_label);
        else:
            return 'Full';

    @staticmethod
    def _versionGroupString(version_group=None):
        if (version_group is None):
            version_group = 'Original';
        return version_group;

    @staticmethod
    def _getVersionLabelDirName(version_label=None):
        if(isinstance(version_label, int)):
            return ("maxheight_{}".format(version_label));
        else:
            return "Full";


    def getVersionPath(self, version_label=None, version_group=None):
        """
        Get path to version of video (path to video file). Return None if version has not been added.
        :param version_label:
        :return:
        """
        return self.getVersionInfo(version_label=version_label, version_group=version_group, info_label='path');

    def getDirForVersion(self, version_label=None, version_group=None):
        vdir = self.getDir('versions') + os.sep + self._versionGroupString(version_group) + os.sep + self._getVersionLabelDirName(version_label)+os.sep;
        return vdir;


    def getVersionInfo(self, version_label, version_group=None, info_label=None):
        """
        Get info about a version of the video
        :param version_label:
        :param info_label:
        :return:
        """
        assert(info_label is not None), "should provide info_label to getVersionInfo()"
        d = self.getVersionDictionary(version_label=version_label, version_group=version_group);
        if(d is not None):
            return d.get(info_label);
        else:
            return None;

    def getVersionDictionary(self, version_label=None, version_group=None):
        if(version_group is None):
            version_group='Original';
        vis_d=self.versions_info.get(version_group);
        if(vis_d is not None):
            return vis_d.get(VideoSource._versionLabelString(version_label));
        else:
            return None;

    def setVersionDictionary(self, version_label=None, version_group=None, d=None):
        if(version_group is None):
            version_group='Original';
        vis_d=self.versions_info.get(version_group);
        if(vis_d is None):
            self.versions_info[version_group]={};
            vis_d=self.versions_info.get(version_group);
        vis_d[VideoSource._versionLabelString(version_label)]=d;
        return;

    def setVersionInfo(self, version_label=None, version_group=None, video_path=None, **kwargs):
        version_dict = self.getVersionDictionary(version_label=version_label, version_group=version_group);
        if(version_dict is None):
            self.setVersionDictionary(version_label=version_label, version_group=version_group, d={});
            version_dict = self.getVersionDictionary(version_label=version_label, version_group=version_group);
        if(video_path is not None):
            version_dict.update({'path':str(pathstring(video_path))});
        if(kwargs is not None):
            version_dict.update(kwargs);


####################

    def hardSave(self):
        if (os.path.isfile(self.getJSONPath())):
            os.rename(self.getJSONPath(), self.getDir('backup') + os.sep + self.AOBJECT_TYPE() + ".json");
        self.writeToJSON(self.getJSONPath());

    def save(self):
        if (self.getInfo('block_writing_to_json')):
            return;
        self.hardSave();

####################


    # def removeAllVersionFiles(self, asset_manager=None):
    #     AWARN("This should wipe the directory, ya?")
    #     # AWARN("Remove all version files does not remove features and vis images! This still needs to be implemented!")
    #     # for vtype in self.versions_info:
    #     #     for v in self.versions_info.get(vtype):
    #     #         # print(self.versions_info.get(vtype).get(v));
    #     #         vpath = self.versions_info.get(vtype).get(v).get('path');
    #     #         assert(vpath), "version {} did not have path".format(vtype)
    #     #         print(vpath)
    #     #         if(os.path.isfile(vpath)):
    #     #             print("REMOVING {}".format(vpath))
    #     #             os.remove(vpath);
    #     #         if(asset_manager is not None and vtype=='Original'):
    #     #             if(v=='Full'):
    #     #                 rs=None;
    #     #             else:
    #     #                 rs = int(v);
    #     #             asset_manager.removeFeaturesForLinkRes(link=self, max_height=rs);
    #     # return True;

    # def removeFeaturesForRes(self, max_height=None):
    #     v = self.getVersionVideo(name=link.name, max_height = max_height);
    #     v.clearFeatureFiles(features_to_clear=['all', 'each']);
    #
    #
    # def removeFeatureFilesForVideo(self, video, features_to_remove=None):
    #     if(features_to_remove is None):
    #         return;
    #     if(not isinstance(features_to_remove, list)):
    #         features_to_remove=[features_to_remove];
    #     for f in features_to_remove:
    #         self.removeFeatureFileForVideo(video=video, feature_name=f);
    #
    #
    # def removeFeatureFileForVideo(self, video, feature_name):
    #     AWARN("still need to implement removeFeatureFileForVideo. Should just empty corresponding directory. Also implement 'each' that removes all.")
    #     # if(feature_name=='each'):
    #     #     return self.removeFeatureFilesForVideo(video=video, features_to_remove=video.getFeatureFunctionsList());
    #     # max_height = video.getInfo(label='max_height');
    #     # link = self.hasLinkWithName(name=video.name);
    #     # source_type=video.getFeatureSohahurceType(feature_name);
    #     # ipath = self.getFeaturePathForLink(feature_name=feature_name, link=link, max_height=max_height, source_type=source_type);
    #     # if(os.path.isfile(ipath)):
    #     #     print("REMOVING {}".format(ipath));
    #     #     os.remove(ipath);
    #
    # def removeFeatureFiles(self, video, feature_name):
    #     if(feature_name=='each'):
    #         AWARN("IMPLEMENT DELETE FEATURE DIR AND CREATE CLEAN");
    #     version_label = video.getInfo(label='version_label');
    #     AWARN("IMPLEMENT DELETE {} FOR VERSION_LABEL {}".format(feature_name, version_label));
    #     # ipath = self.getFeaturePathForLink(feature_name=feature_name, link=link, max_height=max_height, source_type=source_type);
    #     # if(os.path.isfile(ipath)):
    #     #     print("REMOVING {}".format(ipath));
    #     #     os.remove(ipath);
    #
    # def removeVersion(self, version_label=None, remove_files=True):
    #     AWARN("Remove Resolution does not remove features! This still needs to be implemented!")
    #     for vtype in self.versions_info:
    #         v = self.versions_info.get(vtype).get(VideoSource._versionLabelString(version_label));
    #         if(v is not None):
    #             vpath = v.get('path');
    #             if(os.path.isfile(vpath)):
    #                 print("REMOVING {}".format(vpath))
    #                 os.remove(vpath);
    #             self.versions_info.get(vtype).pop(VideoSource._versionLabelString(version_label), None);
    #     return True;
    #
    # def copyResolutionTo(self, version_label=None, output_dir=None):
    #     assert(output_dir)
    #     for vtype in self.versions_info:
    #         v = self.versions_info.get(vtype).get(VideoSource._versionLabelString(version_label));
    #         if(v is not None):
    #             vpath = v.get('path');
    #             if(os.path.isfile(vpath)):
    #                 opath = os.path.join(output_dir,(vtype+'_'+VideoSource._versionLabelString(version_label)+self.video_file_name));
    #                 print("COPYING {} to {}".format(vpath,opath));
    #                 shutil.copy2(vpath, opath);
    #     return True;
    #
    # def copyAssetsTo(self, output_dir=None, asset_manager=None):
    #     for vtype in self.versions_info:
    #         output_type_dir=pathstring(output_dir+os.sep+vtype+os.sep);
    #         make_sure_dir_exists(output_type_dir);
    #         for v in self.versions_info.get(vtype):
    #             output_a_dir = pathstring(output_type_dir+os.sep+v+os.sep);
    #             make_sure_dir_exists(output_a_dir);
    #             # print(self.versions_info.get(vtype).get(v));
    #             vpath = self.versions_info.get(vtype).get(v).get('path');
    #             if(vpath and os.path.isfile(vpath)):
    #                 print("copying {} to {}".format(vpath, output_a_dir));
    #                 shutil.copy2(vpath, output_a_dir);
    #                 if(asset_manager is not None and vtype=='Original'):
    #                     if(v=='Full'):
    #                         rs=None;
    #                     else:
    #                         rs = int(v);
    #                     asset_manager.saveFeaturesForVersion(version_label=rs, output_dir = output_a_dir);

    def getWarpsDir(self, version_label=None):
        warpdir = self.getDir('warps');
        resdir = self._getVersionLabelDirName(version_label=version_label);
        rdir = pathstring(warpdir + os.sep + resdir + os.sep);
        make_sure_path_exists(rdir);
        return rdir;

    # ############################# FROM ASSETMANAGER #######################

    def getVersion(self, max_height=None, get_if_missing=True, load_features=True, **kwargs):
        """
        Gets a version of the video with maximum height max_height. This function contains logic to decide whether/when
            to pull the video. Use pullVersion to force pulling that overwrite's existing assets.
        :param max_height:
        :param get_if_missing:
        :return:
        """
        vpath = self.getVersionPath(version_label=max_height);
        if(vpath and os.path.isfile(vpath)):
            num_frames_total = self.getVersionInfo(version_label=max_height, info_label='num_frames_total');
            if(num_frames_total):
                v = self.RegisteredVideo(path=vpath, version_label = max_height, num_frames_total=num_frames_total, load_features=True);
            else:
                v = self.RegisteredVideo(path=vpath, version_label = max_height, load_features=True);
                self.setVersionInfo(version_label=max_height, num_frames_total=v.num_frames_total);
                self.save();
            return v;
        else:

            if(get_if_missing):
                vpath = self.pullVersion(max_height=max_height, **kwargs);
                if(vpath):
                    num_frames_total = self.getVersionInfo(version_label=max_height, info_label='num_frames_total');
                    if(num_frames_total):
                        v = self.RegisteredVideo(path=vpath, version_label = max_height, num_frames_total=num_frames_total, load_features=True);
                    else:
                        v = self.RegisteredVideo(path=vpath, version_label=max_height, load_features=True);
                        self.setVersionInfo(version_label=max_height, num_frames_total=v.num_frames_total);
                        self.save();
                else:
                    AWARN("COULDNT GET VIDEO FROM {}".format(self.source_location));
            else:
                AWARN("COULDNT FIND VIDEO LOCALLY");
                return;
        return v;

    def saveFeaturesForVideo(self, video, features_to_save=None, output_dir=None, overwrite=True):
        if(features_to_save is None):
            return;
        if(not isinstance(features_to_save, list)):
            features_to_save=[features_to_save];
        for f in features_to_save:
            self.saveFeatureForVideo(video=video, feature_name=f, output_dir=output_dir, overwrite=overwrite);

    def saveFeatureForVideo(self, video, feature_name, output_dir=None, overwrite=True):
        if(feature_name=='each'):
            return self.saveFeaturesForVideo(video=video, features_to_save=video.getFeaturesList(), output_dir=output_dir, overwrite=overwrite);

        version_label = video.getInfo(label='version_label');
        source_type = video.getFeatureSourceType(feature_name);
        opath = self.getFeaturePath(feature_name=feature_name, version_label=version_label, source_type=source_type, output_dir=output_dir);
        make_sure_dir_exists(opath);
        if(not os.path.isfile(opath) or overwrite):
            if(feature_name=='all'):
                video.saveFeatures(path=opath);
                return;
            else:
                vfeature = video.getFeature(name=feature_name, force_recompute=False);
                if(vfeature is not None):
                    video.saveFeature(name=feature_name, path=opath);

    def loadFeaturesForVideo(self, video, features_to_load=None):
        if(features_to_load is None):
            return;
        if(not isinstance(features_to_load, list)):
            features_to_load=[features_to_load];
        for f in features_to_load:
            self.loadFeatureForVideo(video=video, feature_name=f);

    def loadFeatureForVideo(self, video, feature_name):
        if(feature_name=='each'):
            return self.loadFeaturesForVideo(video=video, features_to_load=video.getFeatureFunctionsList());
        version_label = video.getInfo(label='version_label');
        source_type=video.getFeatureSourceType(feature_name);
        ipath = self.getFeaturePath(feature_name=feature_name, version_label=version_label, source_type=source_type);
        if(os.path.isfile(ipath)):
            if(feature_name=='all'):
                video.loadFeatures(path=ipath);
            else:
                video.loadFeature(name=feature_name, path=ipath);

    def getFeaturePath(self, feature_name=None, source_type=None, version_label=None, output_dir=None):
        assert(feature_name is not None), 'must provide name of feature VideoSource.getFeaturePath'
        feature_dir = self.getFeatureDir(feature_name=feature_name, source_type = source_type);
        fileext = '.pkl';
        if(output_dir is None):
            version = self._getVersionLabelDirName(version_label=version_label)+os.sep;
            output_dir = pathstring(feature_dir);
        outname = self.getName();
        if(outname is None):
            outname = 'version';
        outname = outname+'_'+self._getVersionLabelDirName(version_label=version_label)+fileext;
        opath = os.path.join(output_dir, outname);
        return opath;

    def getFeatureDir(self, feature_name=None, source_type=None):
        if(source_type is None):
            source_type='video'; # could be 'audio'
        # AWARN("getDir('features')= {}\nsource_type= {}\nfeature_name= {}".format(self.getDir('features'), source_type, feature_name))
        ftypedir = pathstring(self.getDir('features') + os.sep + source_type + os.sep + feature_name + os.sep);
        return ftypedir;

    def saveFeaturesForVersion(self, version_label=None, output_dir=None):
        assert(output_dir), 'must provide output dir';
        v = self.getVersion(max_height=version_label);
        v.save(features_to_save='all', output_dir=output_dir);


    def RegisteredVideo(self, path=None, version_label=None, version_group = None, num_frames_total=None, load_features = True):
        v = self.VideoClass(path=path, name=self.getName()+'_'+self._versionLabelString(version_label=version_label), num_frames_total=num_frames_total);
        # v = Video(path=path, name=self.getName()+'_'+self._versionLabelString(version_label=version_label), num_frames_total=num_frames_total);
        self.RegisterVideo(video=v, version_label=version_label, load_features=load_features);
        return v;

    def RegisterVideo(self, video, version_label = None, version_group = None, load_features=True):
        def videosave(vidob, features_to_save='all', output_dir=None, overwrite=True):
            self.saveFeaturesForVideo(video=vidob, features_to_save=features_to_save, output_dir=output_dir, overwrite=overwrite);
            #self.save
        def videoload(vidob, features_to_load='all', input_dir=None):
            self.loadFeaturesForVideo(video=vidob, features_to_load=features_to_load);
        def videoclear(vidob, features_to_clear='all'):
            self.removeFeatureFilesForVideo(video=vidob, features_to_remove=features_to_clear);

        video.setInfo(label='version_label', value=version_label);
        video.setInfo(label='version_group', value=version_group);
        video.setInfo(label='video_file_name', value=os.path.split(video.getPath())[1]);
        # if(version_group is None):
        video.save_func = videosave;
        video.load_func = videoload;
        video.clear_feature_files_func = videoclear;
        video.source=self;

        if(load_features is True and hasattr(video, "loadFlowFeatures")):
            video.loadFlowFeatures();
        elif(load_features is not None and (isinstance(load_features, list) or isinstance(load_features, str))):
            video.load(features_to_load = load_features);

        # return video;

    def addVersion(self, path, version_label=None, version_group=None):
        v = self.RegisteredVideo(path=path, version_label=version_label);
        video_dict=v.getVersionInfo();
        video_dict.update(dict(version_group=version_group));
        self.setVersionInfo(video_path = path, version_label=version_label, **video_dict);
        self.save();

    def addVersionToVideo(self, video, new_video, version_label=None, version_group=None):
        version_label = video.getInfo(label='version_label');
        video_dict = new_video.getVersionInfo();
        if ((version_group is not None) or (video_dict.get('version_group') is None)):
            video_dict.update(dict(version_group=version_group));
        self.setVersionInfo(video_path=new_video.getPath(), version_label=version_label, **video_dict);
        self.save();


    def setSource(self, source_location=None, assert_valid=True, **kwargs):
        """
        Sets the source of this VideoSource. If file, copies the file to VideoSource directory as "Original" version,
            unless copy argument is set to false.
        :param source_location: either path to file, or youtube url
        :param kwargs: see setSourceYoutube or setSourceFile for options
        :return:
        """

        if(source_location):
            if(os.path.isfile(source_location)):
                self.setSourceFile(path = source_location, **kwargs);
            else:
                self.setSourceYoutube(url=source_location, **kwargs);
            return;
        elif(kwargs.get('video_path')):
            self.setSourceFile(**kwargs);
            return;
        elif(kwargs.get('url')):
            self.setSourceYoutube(**kwargs);
            return;
        else:
            if(assert_valid):
                assert(False),'No valid source location provided to setSource.'




    def setSourceYoutube(self, url=None, max_height=None, pull_fullres=True, **kwargs):
        self.source_location=url;
        self.setInfo(label='source_type', value='youtube');
        if(pull_fullres):
            self.pullYoutubeVideo(max_height=max_height, **kwargs);
        self.save();

    def setSourceFile(self, path=None, copy=True, **kwargs):
        # assert(os.path.isfile(path)), 'Tried to set source but no file exists at {}'.format(path);
        if(not os.path.isfile(path)):
            return None;

        self.video_file_name = os.path.basename(path);
        if(not copy):
            self.source_location = path;
            self.setInfo(label='source_type', value='file_address');
            return path;
        # output_dir = self.getDir('versions') + os.sep + 'Original' + os.sep;
        output_dir = self.getDirForVersion(version_label=None, version_group='Source');
        make_sure_dir_exists(output_dir);
        output_path = os.path.join(output_dir, self.video_file_name);
        # max_height = kwargs.get('max_height');

        # if(max_height is not None):
        #     # print('max_height was {}'.format(max_height));
        #     original = self.VideoClass(path=path);
        #     original.writeResolutionCopyFFMPEG(path=output_path, max_height=kwargs.get('max_height'));
        # else:
        #     shutil.copy2(path, output_path);

        shutil.copy2(path, output_path);

        # self.addVersion(path=output_path, version_label='Original');
        self.addVersion(path=output_path, version_label='Full');
        self.setInfo(label='source_type', value='file');
        self.source_location = output_path;
        self.save();
        return output_path;


    def pullVersion(self, max_height=None, **kwargs):
        # assert(self.source_location is not None), "Could not pull version; source location is None."
        if(self.source_location is None):
            return None;
        source_type = self.getInfo('source_type');
        if(source_type=='youtube'):
            return self.pullYoutubeVideo(max_height=max_height, **kwargs);
        else:
            return self.pullFileVideo(max_height=max_height, **kwargs)


    def pullFileVideo(self, max_height=None, force_recompute=None):
        original_path = self.source_location;
        assert(os.path.isfile(original_path)), 'SOURCE FILE {} IS MISSING'.format(self.source_location);
        original = self.VideoClass(path=original_path);
        output_dir = self.getDirForVersion(version_label=max_height, version_group=None);
        make_sure_dir_exists(output_dir);
        output_path = os.path.join(output_dir, self.video_file_name);
        if(max_height is None):
            if(os.path.normpath(original_path) != os.path.normpath(output_path)):
                shutil.copy2(original_path, output_path);
        else:
            original.writeResolutionCopyFFMPEG(path=output_path, max_height=max_height);
        self.addVersion(path=output_path, version_label=max_height);
        return self.getVersionPath(version_label=max_height);

    def pullYoutubeVideo(self, max_height=None, write_subtitles=False, save_youtube_info=False,
                  force_redownload=False):
        """
        Downloads video from youtube with height<=max_height. Returns path to downloaded video.
        :param max_height: maximum height of video to download
        :param write_subtitles: whether to save the subtitles
        :param save_youtube_info: whether to save the info json
        :param force_redownload: whether to re-download if video exists
        :return:
        """
        assert (not self.getInfo('source_is_file')), "tried to call pullYoutubeVideo on file source {}.".format(self.getName());

        old_vid_path = self.getVersionPath(version_label=max_height);

        if ((not force_redownload) and old_vid_path and os.path.isfile(old_vid_path)):
            AWARN("Old video version exists with path {}\nSkipping download...\n".format(old_vid_path))
            return old_vid_path;

        # output_dir = self.getDir('versions') + os.sep + self._getVersionLabelDirName(version_label=max_height) + os.sep;
        output_dir = self.getDirForVersion(version_label=max_height, version_group=None);
        make_sure_path_exists(output_dir);

        if ((not force_redownload) and self.video_file_name is not None):
            path_guess = os.path.join(output_dir, self.video_file_name);
            if (os.path.isfile(path_guess)):
                AWARN("found existing file {}\nSkipping download; set force_redownload=True to overwrite old version.".format(path_guess));
                self.addVersion(path=path_guess, version_label=max_height);
                return path_guess;


        ########Download From YouTube#########
        vidpath_template = output_dir + '%(title)s-%(id)s' + '.%(ext)s';
        ydl_opts = {
            'format': 'bestvideo[ext=mp4]+bestaudio[ext=m4a]/best[ext=mp4]/best',
            'outtmpl': vidpath_template,
        }
        if (write_subtitles):
            ydl_opts['writesubtitles'] = True;
            ydl_opts['subtitlesformat'] = '[srt]';
        if (max_height):
            ydl_opts['format'] = 'bestvideo[height<={}][ext=mp4]+bestaudio[ext=m4a]/best[ext=mp4]/best'.format(
                max_height);
        ydl_opts['writeinfojson'] = True;
        info_dict = None;

        AWARN("Downloading {} from {}...".format(self.getName(), self.source_location));

        with youtube_dl.YoutubeDL(ydl_opts) as ydl:
            info_dict = ydl.extract_info(self.source_location, download=True)
            # video_source = info_dict.get("source", None)
            # video_id = info_dict.get("id", None)
            # video_title = info_dict.get('title', None)
        usedtitle = info_dict['title'].replace('"', "'").replace('|', '_').replace(':', ' -').replace('/','_').replace('?', '');
        usedfilename_withoutext = usedtitle + '-' + info_dict['id'];
        usedfilename = usedfilename_withoutext + '.' + info_dict['ext'];
        filepath = output_dir + os.sep + usedfilename;
        filepathsafe = safe_file_name(filepath);
        if (os.path.isfile(filepath)):
            os.rename(filepath, filepathsafe);
        else:
            fpath = glob.glob(output_dir + os.sep + '*' + info_dict['id'] + '.mp4');
            assert (len(fpath) == 1), "Wrong number of files for {}\nFound:\n{}".format(filepath, fpath);
            os.rename(fpath[0], filepathsafe);
            jpath = glob.glob(output_dir + os.sep + '*' + info_dict['id'] + '.info.json');
            os.rename(jpath[0], safe_file_name(usedfilename_withoutext + '.info.json'));

        print(("Saved to {}".format(filepathsafe)));

        self.video_file_name = safe_file_name(usedfilename);
        self.title_safe = safe_file_name(usedtitle);
        if (save_youtube_info):
            self.youtube_info_dict = info_dict;

        # v = Video(path=filepathsafe);
        # video_dict=v.toDictionary();
        # self.setVersionInfo(video_path = filepathsafe, version_label=max_height, num_frames_total=v.num_frames_total, **video_dict);
        # self.save();
        self.addVersion(path=filepathsafe, version_label=max_height);
        return self.getVersionPath(version_label=max_height);
        # return youtube_link;



from .VBObject import *
from .Event import *
import math
from operator import truediv
from .VisBeatImports import *

class TimeSignal(VBObject):
    """TimeSignal (class): A time signal, and a bunch of convenience functions to go with it.
        Attributes:
            sampling_rate: the sampling rate
            visualizations dictionary of visualizations. (removed - too many dependencies)
    """

    def VBBJECT_TYPE(self):
        return 'TimeSignal';

    def __init__(self, path=None, sampling_rate = None):
        VBObject.__init__(self, path=path);
        # self.initializeBlank();
        # self.visualizations = AFuncDict(owner=self, name='visualizations');
        # self.visualizations.functions.update(self.VIS_FUNCS);
        if(sampling_rate):
            self.sampling_rate=sampling_rate;

    def initializeBlank(self):
        VBObject.initializeBlank(self);
        self.sampling_rate = None;


    # <editor-fold desc="Property: 'frame_rate'">
    @property
    def frame_rate(self):
        return self._getFrameRate();
    def _getFrameRate(self):
        raise NotImplementedError;
    # </editor-fold>


    def getSampleAtTime(self, f):
        prev_sample = self.getSampleAtIndex(math.floor(f));
        next_sample = self.getSampleAtIndex(math.ceil(f));
        sample_progress = f-np.floor(f);
        return (next_sample*sample_progress)+(prev_sample*(1.0-sample_progress));

    def getSampleAtIndex(self, i):
        return self.getTimeForIndex();

    def getDuration(self):
        assert(False), "getDuration must be implemented for subclass of TimeSignal"

    def getSampleDuration(self):
        return 1.0/self.sampling_rate;

    def getTimeForIndex(self, i):
        return i*self.getSampleDuration();

    # def toDictionary(self):
    #     d = VBObject.toDictionary(self);
    #     assert(False), "haven't implemented toDictionary for {} yet".format(self.VBOBJECT_TYPE())
    #     #serialize class specific members
    #     return d;

    # def initFromDictionary(self, d):
    #     VBObject.initFromDictionary(self, d);
    #     assert(False), "haven't implemented initFromDictionary for {} yet".format(self.VBOBJECT_TYPE())
    #     #do class specific inits with d;



from .AObject import *
#import shutil
from distutils.dir_util import copy_tree

class AFileManager(AObject):
    """AFileManager (class): Manages assets. This should really be replaced with a database of some sort...
        Attributes:
            todo
    """

    @staticmethod
    def AOBJECT_TYPE():
        return 'AFileManager';

    def getJSONPath(self):
        return self.getPath();

    def __init__(self, path=None, clear_temp=None):
        """If you provide a directory, it will look for a existing AFileManager.json in that directory, or create one if it does not already exist.
            If you provide a json, it will use that json, unless the json doesn't exist, in which case it will complain...
        """
        AObject.__init__(self, path=path);
        # self.initializeBlank();
        self.initWithPath(path=path, clear_temp=clear_temp);

    def initializeBlank(self):
        AObject.initializeBlank(self);
        self.directories = {};

    def getJSONName(self):
        return self.AOBJECT_TYPE()+".json";

    def initWithPath(self, path=None, clear_temp=None):
        oldpath = None
        newpath = path;
        if(path):
            if(os.path.isfile(path)):
                self.loadFromJSON(self.getJSONPath()); #assume path property is already set to 'path'
                oldpath = self.getPath(); #whatever was in the json, having overwritten path property
            elif(os.path.isdir(path)):
                json_file_path = path+os.sep+self.getJSONName();
                self.setPath(json_file_path);
                if(os.path.isfile(self.getJSONPath())):
                    self.loadFromJSON(json_file_path);
                    oldpath = self.getPath();
                    newpath = json_file_path;
                    # self.setPath(file_path=json_file_path);
                else:
                    newpath=self.getJSONPath()
                    self.writeToJSON(json_path=newpath);#no json file found, so we create one
            else:
                assert False, "Given AFileManager path is neither an existing directory or file! path: {} (AFileManager.py)".format(path)

            self.setPath(file_path=newpath);

            if(oldpath):
                oldir = get_dir_from_path(pathstring(oldpath));
                newdir = get_dir_from_path(pathstring(newpath));
                if(oldir != newdir):
                    AWARN("FILEMANAGER FOUND FILE MOVED FROM:\n{}\nTO:\n{}\nUPDATING DIRECTORIES...".format(oldir,
                                                                                                            newdir));
                    for d in self.directories:
                        dpth = self.directories[d];
                        if(dpth.startswith(oldir)):
                            dpthst = dpth.lstrip(oldir);
                            self.directories[d]=os.path.join(newdir,dpthst);
                            AWARN("{} updated to {}".format(dpth, self.directories[d]));


            self.setDir('data', pathstring(self.getDirectoryPath()+os.sep+"Data"+os.sep));
            self.setDir('backup', pathstring(self.getDir('data')+"Backups"+os.sep));
            self.setDir('temp', pathstring(self.getDir('data')+"TEMP"+os.sep));
            temp_dir = self.getDir('temp');
            if(os.path.isdir(temp_dir) and (clear_temp)):
                for the_file in os.listdir(temp_dir):
                    file_path = os.path.join(temp_dir, the_file);
                    try:
                        if os.path.isfile(file_path):
                            os.remove(file_path);
                            #os.unlink(file_path);
                            #elif os.path.isdir(file_path): shutil.rmtree(file_path)
                    except Exception as e:
                        print(e)
            make_sure_path_exists(temp_dir);
            #Video.VIDEO_TEMP_DIR = temp_dir;

    def setDir(self, name, path):
        # AWARN("setting {} to {}".format(name, path))
        # assert(name is not 'log')
        self.directories[name]=path;
        make_sure_path_exists(path);
        return path;

    def addDir(self, name):
        assert(name not in self.directories), "tried to add {} dir to AFileManager, but this dir is already set"
        return self.setDir(name, pathstring(self.getDirectoryPath()+os.sep+name+os.sep));

    def getDir(self, name):
        # printDictionary(self.directories)
        return self.directories.get(name);


    def emptyDir(self, name):
        dpth = self.getDir(name);
        if(dpth is not None and os.path.isdir(dpth)):
            shutil.rmtree(dpth);
            make_sure_path_exists(dpth);

    def deleteDir(self, name):
        dpth = self.getDir(name);
        if (dpth is not None and os.path.isdir(dpth)):
            shutil.rmtree(dpth);
            d = dict(self.directories);
            del d[name];
            self.directories=d;


    def toDictionary(self):
        d = AObject.toDictionary(self);
        d['directories']=self.directories;
        #serialize class specific members
        return d;

    def copyPathToDir(self, path_to_copy, dest_dir):
        dest_path = self.getDir(dest_dir);
        if(dest_path):
            if(os.path.isdir(path_to_copy)):
                copy_tree(src=path_to_copy, dst=dest_path);
            elif(os.path.isfile(path_to_copy)):
                shutil.copy2(path_to_copy, dest_path)
        return;

    def copyDirToPath(self, dir_to_copy, dest_path):
        src_path = self.getDir(dir_to_copy);
        if(src_path):
            if(os.path.isdir(dest_path)):
                copy_tree(src=src_path, dst=dest_path);
        return;

    @staticmethod
    def copyRandomFractionOfFilesInSourceDir(source_dir, dest_dir, fraction=1.0, ext=None):
        """
        Copies a random fraction of files in source directory... Wrote this for splitting training/test data in ML applications.
        :param source_dir:
        :param dest_dir:
        :param fraction:
        :param ext:
        :return:
        """
        directories = []
        subdirnames = []
        filepaths = [];
        for filename in os.listdir(source_dir):
            path = os.path.join(source_dir, filename)
            if os.path.isdir(path):
                directories.append(path)
                subdirnames.append(filename)
            else:
                # namepart, extpart = os.path.splitext(filename);
                if((ext is None) or filename.lower().endswith(ext)):
                    filepaths.append(path);

        n_to_copy = int(len(filepaths)*fraction);
        random_seed = 0;
        random.seed(random_seed);
        random.shuffle(filepaths);
        copy_sources = filepaths[:n_to_copy];
        for src, dst in zip(copy_sources, [dest_dir]*len(copy_sources)):
            #print("src: {}\ndst: {}".format(src, dst));
            shutil.copy2(src, dst);

        for d in range(len(directories)):
            subdest = pathstring(os.path.join(dest_dir,subdirnames[d])+os.sep);
            make_sure_dir_exists(subdest);
            AFileManager.copyRandomFractionOfFilesInSourceDir(source_dir=directories[d], dest_dir=subdest, fraction=fraction, ext=ext);

    def initFromDictionary(self, d):
        AObject.initFromDictionary(self, d);
        self.directories = d['directories'];

    def save(self):
        if(os.path.isfile(self.getJSONPath())):
            os.rename(self.getJSONPath(), self.getDir('backup')+os.sep+self.AOBJECT_TYPE()+".json");
        self.writeToJSON(self.getJSONPath());



from .VisBeatExampleVideo import *
dances = [];

music = [];

music.append(VisBeatExampleVideo(display_name = 'U Cant Touch This', name='cant_touch_this', url='https://www.youtube.com/watch?v=_NNYI8VbFyY', start_beat=2, leadin=2));
music.append(VisBeatExampleVideo(display_name = 'Mary Poppins: Supercalifrag.', name='supercalifrag', url='https://www.youtube.com/watch?v=rihNRTTcztQ', start_beat = 5, leadin = 3));
music.append(VisBeatExampleVideo(display_name = 'Migos: Walk It Talk It',name='walkit_talkit', url='https://www.youtube.com/watch?v=fGqdIPer-ms', start_beat = 134));
music.append(VisBeatExampleVideo(display_name = 'Rammstein: Du Hast',name='du_hast', url = 'https://www.youtube.com/watch?v=W3q8Od5qJio', start_beat = 193));
music.append(VisBeatExampleVideo(display_name = 'Taylor Swift: Shake It Off', name='shake_it_off', url='https://www.youtube.com/watch?v=nfWlot6h_JM', start_beat=112));
music.append(VisBeatExampleVideo(display_name = 'Michael Jackson: Billie Jean', name='billie_jean', url='https://www.youtube.com/watch?v=Zi_XLOBDo_Y', start_beat=169));
music.append(VisBeatExampleVideo(display_name = 'Hall and Oates: You Make My Dreams', name='you_make_my_dreams', url='https://www.youtube.com/watch?v=EErSKhC0CZs', start_beat=95));
music.append(VisBeatExampleVideo(display_name = 'Parry Gripp: Breakfast Burrito',name="breakfast_burrito", url='https://www.youtube.com/watch?v=prPjpwsGiws', start_beat=32))
music.append(VisBeatExampleVideo(display_name = 'Pharrell Williams: Happy',name='happy', url='https://www.youtube.com/watch?v=y6Sxv-sUYtM', start_beat=69, leadin=3))
music.append(VisBeatExampleVideo(display_name = 'Aqua: Barbie Girl',name='barbie_girl', url='https://www.youtube.com/watch?v=ZyhrYis509A', start_beat=67, leadin=4));
music.append(VisBeatExampleVideo(display_name = 'Jhameel: Montage',name='music_cc_jhameel_montage', url='https://www.youtube.com/watch?v=vYr_nCmsgoE', start_beat=103))
music.append(VisBeatExampleVideo(display_name = 'Vengaboys: We like to Party',name='we_like_to_party', url='https://www.youtube.com/watch?v=6Zbi0XmGtMw', start_beat=65))
music.append(VisBeatExampleVideo(display_name = 'Rick Astley: Never Gonna Give You Up', name='rick_roll', url='https://www.youtube.com/watch?v=dQw4w9WgXcQ', start_beat = 3, leadin=3));
music.append(VisBeatExampleVideo(display_name = 'Mark Ronson: Uptown Funk ft. Bruno Mars', name='uptown_funk', url='https://www.youtube.com/watch?v=OPf0YbXqDm0', start_beat=32));
music.append(VisBeatExampleVideo(display_name = 'Missy Elliot: Work It', name='work_it', url='https://www.youtube.com/watch?v=cjIvu7e6Wq8', start_beat=52, leadin = 1));
music.append(VisBeatExampleVideo(display_name = 'If I only had a brain', name='only_had_a_brain', url='https://www.youtube.com/watch?v=nauLgZISozs', start_beat=14));
music.append(VisBeatExampleVideo(display_name = 'Kane Brown: Lose It', name='kb_lose_it', url='https://www.youtube.com/watch?v=Z8nkc2KKjd8', start_beat=56));
# music.append(VisBeatExampleVideo(display_name = 'Check Meowt', name='pg_check_meowt', url='https://youtu.be/NjKYX4bQpf0', start_beat = 16));
# music.append(VisBeatExampleVideo(display_name = 'Wu-Tang Clan: Protect Ya Neck', name='wutang_protect_ya_neck', url='https://www.youtube.com/watch?v=R0IUR4gkPIE', start_beat = None));
# music.append(VisBeatExampleVideo(display_name = 'Ice Cube: It Was A Good Day',name='it_was_a_good_day', url='https://www.youtube.com/watch?v=h4UqMyldS7Q', start_beat=64, harmonic = 'half'));


from .EventList import *
from .TimeSignal1D import *
from scipy.io.wavfile import write
import librosa
import librosa.display

from scipy import signal, fftpack



class Audio(TimeSignal1D):
    """Audio (class): A sound, and a bunch of convenience functions to go with it.
        Attributes:
            x: the sound signal
            sampling_rate: the sampling rate
    """

    FEATURE_FUNCS = TimeSignal1D.FEATURE_FUNCS.copy();

    def __init__(self, path=None, sampling_rate=None, x=None, name=None):
        #VBObject.__init__(self, path=path);
        TimeSignal1D.__init__(self, path=path, sampling_rate=sampling_rate, x=x);
        #self.initializeBlank();
        if(name is not None):
            self.name = name;
        if(path):
            self.loadFile();
        if(self.name is None):
            self.name = self.getInfo('file_name')


    # <editor-fold desc="Property: 'name'">
    @property
    def name(self):
        return self.getName();
    def getName(self):
        return self._name;
    @name.setter
    def name(self, value):
        self._setName(value);
    def _setName(self, value):
        self._name = value;
    # </editor-fold>

    def initializeBlank(self):
        self.name = None;
        TimeSignal1D.initializeBlank(self);
        self.n_channels = 1;

    def _getFrameRate(self):
        return self.getOnsetSamplingRate();

    def clone(self):
        clone = Audio();
        clone.setPath(self.getPath());
        clone.x = self.x.copy();
        clone.sampling_rate = self.sampling_rate;
        clone.n_channels = self.n_channels;
        stereo = self.getStereo();
        if (stereo is not None):
            clone.setInfo('stereo_signal', stereo);
            clone.setInfo('stereo_sampling_rate', self.getInfo('stereo_sampling_rate'));
        return clone;

    def loadFile(self, file_path=None, sampling_rate=None, convert_to_mono=True):
        if(file_path):
            self.setPath(file_path=file_path);

        if('file_path' in self.a_info):
            self.x, self.sampling_rate = librosa.load(self.a_info['file_path'],sr=sampling_rate, mono=convert_to_mono);
            if(len(self.x.shape)>1):
                self.a_info['stereo_signal']=self.x;
                self.setInfo('stereo_sampling_rate', self.sampling_rate);
                self.x = np.mean(self.x, axis = 0)
                print("averaged stereo channels");

    def getStereo(self):
        return self.a_info.get('stereo_signal');

    def getStereoSamplingRate(self):
        return self.getInfo('stereo_sampling_rate');



    def getStringForHTMLStreamingBase64(self):
        encoded = self.getStereoEncodedBase64WAV();
        return "data:audio/wav;base64,{0}".format(encoded.decode('ascii'));

    def getStereoEncodedBase64WAV(self):
        sig = self.getStereo();
        sr = self.getStereoSamplingRate();
        if (sig is None):
            sig = self.getSignal();
            sr = self.sampling_rate;
        saudio = _make_wav(sig, sr);
        encoded = base64.b64encode(saudio);
        return encoded;

    def getMonoEncodedBase64WAV(self):
        sig = self.getSignal();
        sr = self.sampling_rate;
        saudio = _make_wav(sig, sr);
        encoded = base64.b64encode(saudio);
        return encoded;

    def getLocalRhythmicSaliency(self, **kwargs):
        return self.getOnsetEnvelope(**kwargs);


    def play(self, autoplay = None):
        """
         Play audio. Works in Jupyter. if notebook, audio will play normalized -- this is something html5 audio seems to do by default.
        :param autoplay:
        :return:
        """
        if(ISNOTEBOOK):
            # if(normalize):
            audiodisp = vb_get_ipython().display.Audio(data=self.getSignal(), rate=self.sampling_rate, autoplay=autoplay);
            vb_get_ipython().display.display(audiodisp);
            # vb_get_ipython().display.display(vb_get_ipython().display.Audio(data=self.getSignal(), rate=self.sampling_rate, autoplay=False))
            # else:
            # audiodisp = vb_get_ipython().display.Audio(data=self.getMonoEncodedBase64WAV(), rate=self.sampling_rate,
            #                                            autoplay=autoplay);
            # vb_get_ipython().display.display(audiodisp);
            # print("html render")
            # htmlstr = self.getStringForHTMLStreamingBase64()
            # ahtml = HTML(data='''<audio alt="{}" controls>
            #             <source src="{}" type="audio/wav" />
            #          </audio>'''.format(self.name, htmlstr));
            # IPython.display.display(ahtml);

        else:
            p = pyaudio.PyAudio()
            stream = p.open(format=pyaudio.paFloat32,
                             channels=self.n_channels,
                             rate=self.sampling_rate,
                             output=True,
                             output_device_index=1
                             )
            stream.write(self.getSignal())
            stream.stop_stream();
            stream.close()
            p.terminate()

    def playBeats(self, indices=None, beats=None):
        beat_inds = indices;
        if(beats is None):
            beats = self.getBeatEvents();
        if(beat_inds is None):
            beat_inds = [0, len(beats)-1];
        if(not isinstance(beat_inds, list)):
            beat_inds=[beat_inds-1, beat_inds, beat_inds+1];
        if(beat_inds[0]<0):
            start_time = 0;
        else:
            start_time = beats[beat_inds[0]].start;
        if(beat_inds[-1]>len(beats)):
            end_time = self.getDuration();
        else:
            end_time = beats[beat_inds[-1]].start;
        self.playSegment(time_range = [start_time, end_time]);

    def AudioClipFromBeatRange(self, beat_range, beats=None):
        if(beats is None):
            beats = self.getBeatEvents();
        if(beat_range is None):
            beat_range = [0, len(beats)-1];
        if(beat_range[1] is None):
            beat_range[1]=len(beats)-1;

        return self.AudioClip(start=beats[beat_range[0]].start, end=beats[beat_range[1]].start);




    def playSegment(self, time_range, autoplay=None):
        start_time = time_range[0];
        end_time = time_range[1];
        if(isinstance(start_time, Event)):
            start_time = start_time.start;
        if(isinstance(end_time, Event)):
            end_time = end_time.start;
        if(ISNOTEBOOK):
            audiodisp = vb_get_ipython().display.Audio(data=self.getSignalSegment(time_range=[start_time, end_time]), rate=self.sampling_rate, autoplay=autoplay);
            vb_get_ipython().display.display(audiodisp);
            # vb_get_ipython().display.display(vb_get_ipython().display.Audio(data=self.getSignal(), rate=self.sampling_rate, autoplay=False))
        else:
            p = pyaudio.PyAudio()
            stream = p.open(format=pyaudio.paFloat32,
                             channels=self.n_channels,
                             rate=self.sampling_rate,
                             output=True,
                             output_device_index=1
                             )
            stream.write(self.getSignalSegment(time_range=[start_time, end_time]))
            stream.stop_stream();
            stream.close()
            p.terminate()

    def writeToFile(self, output_path=None, output_sampling_rate=None):
        assert(output_path), "must provide path to save audio."
        data = self.getSignal();
        scaled = np.int16(data/np.max(np.abs(data)) * 32767)
        if(output_sampling_rate is None):
            output_sampling_rate=44100
        write(output_path, output_sampling_rate, scaled)

    def setValueRange(self, value_range=None):
        if(value_range is None):
            value_range = [-1,1];
        TimeSignal1D.setValueRange(self, value_range=value_range);

    def resample(self, sampling_rate):
        new_n_samples = sampling_rate*self.getDuration();
        self.x = sp.signal.resample(self.x, int(new_n_samples));
        self.sampling_rate = sampling_rate;

    def GetResampled(self, sampling_rate):
        new_a = self.clone();
        new_a.resample(sampling_rate=sampling_rate);
        return new_a;



    def AlignedTo(self, B):
        assert(False),'Abe needs to fix -- broke when messing with alignment code for video';

        A = self.clone();
        if(A.sampling_rate !=B.sampling_rate):
            A.resample(B.sampling_rate);

        a_signal = A.getSignal();
        b_signal = B.getSignal();
        a_duration = self.getDuration();

        if(len(a_signal) < len(b_signal)):
            siglen = spotgt_shift_bit_length(len(b_signal));
        else:
            siglen = spotgt_shift_bit_length(len(a_signal));

        npada = siglen-len(a_signal);
        if(npada>0):
            a_signal = np.pad(a_signal, (0,npada), 'constant', constant_values=(0, 0));

        npadb = siglen-len(b_signal);
        if(npadb>0):
            b_signal = np.pad(b_signal, (0,npadb), 'constant', constant_values=(0, 0));

        Af = fftpack.fft(a_signal);
        Bf = fftpack.fft(b_signal);
        Ar = Af.conjugate();
        Br = Bf.conjugate();
        # Ar = -Af.conjugate();
        # Br = -Bf.conjugate();

        ashift = np.argmax(np.abs(fftpack.ifft(Ar * Bf)));
        print(ashift);
        print((np.argmax(np.abs(fftpack.ifft(Af * Br)))));

        a_return = Audio();
        a_return.n_channels = 1;
        a_return.sampling_rate = B.sampling_rate;
        a_return.x = np.roll(a_signal, ashift);

        return a_return;

    def getOffsetFrom(self, B):
        return -self.getShiftAmountTo(B);

    def getShiftAmountTo(self, B):
        """
        Get amount to shift by in seconds (to shift this to alignment with B)
        """
        a_signal = self.getSignal();
        b_signal = B.getSignal();
        a_duration = self.getDuration();
        if(self.sampling_rate !=B.sampling_rate):
            ansamps = a_duration*B.sampling_rate;
            a_signal = sp.signal.resample(a_signal, int(ansamps));
            # a_signal = librosa.resample(a_signal, self.sampling_rate, B.sampling_rate, res_type='kaiser_fast');
        if(len(a_signal) < len(b_signal)):
            siglen = spotgt_shift_bit_length(len(b_signal));
        else:
            siglen = spotgt_shift_bit_length(len(a_signal));
        npada = siglen-len(a_signal);
        if(npada>0):
            a_signal = np.pad(a_signal, (0,npada), 'constant', constant_values=(0, 0));
        npadb = siglen-len(b_signal);
        if(npadb>0):
            b_signal = np.pad(b_signal, (0,npadb), 'constant', constant_values=(0, 0));

        # if(len(a_signal) < len(b_signal)):
        #     npad = len(b_signal)-len(a_signal);
        #     a_signal = np.pad(a_signal, (0,npad), 'constant', constant_values=(0, 0));
        #     print('pad1');
        #
        # if (len(b_signal) < len(a_signal)):
        #     npad = len(a_signal) - len(b_signal);
        #     b_signal = np.pad(b_signal, (0, npad), 'constant', constant_values=(0, 0));
        #     print('pad2');

        # print('point0')

        Af = fftpack.fft(a_signal);
        Bf = fftpack.fft(b_signal);
        Ar = Af.conjugate();
        Br = Bf.conjugate();
        # Ar = -Af.conjugate();
        # Br = -Bf.conjugate();
        ashiftab = np.argmax(np.abs(fftpack.ifft(Ar * Bf)));
        durationsamples = self.getDuration()*B.sampling_rate;
        # if(ashiftab>(durationsamples*0.5)):
        #     ashiftab = ashiftab-durationsamples;
        return truediv(ashiftab, B.sampling_rate);

    def getBeatEventList(self, time_range = None):
        beats = self.getBeats();
        if(time_range is None):
            return EventList.FromStartTimes(beats, type='beats');

        start_beat = 0;
        end_beat = len(beats) - 1;
        if(time_range[0] is not None):
            while((start_beat<len(beats)) and (beats[start_beat]<time_range[0])):
                start_beat=start_beat+1;
        if(time_range[1] is not None):
            while(end_beat>0 and (beats[end_beat]>time_range[1])):
                end_beat = end_beat-1;
        if(end_beat>start_beat):
            return EventList.FromStartTimes(beats[start_beat:end_beat], type='beats');
        else:
            return None;


    def getBeatEvents(self, start_time=None, end_time=None):
        beat_eventlist = self.getBeatEventList();
        return beat_eventlist.events;

    def AudioClip(self, start, end):
        from . import AudioClip
        clip = AudioClip.AudioClip(path=self.getPath(), start=start, end=end);
        return clip;


    def getWithSoundAdded(self, add_times, sound=None, mute_original=None, gain_original = None):
        if(sound is None):
            sound = Audio.PingSound(sampling_rate=self.sampling_rate);
        if(sound.sampling_rate==self.sampling_rate):
            s_toadd = sound.getSignal();
        else:
            new_n_samples = self.sampling_rate * sound.getDuration();
            s_toadd = sp.signal.resample(sound.getSignal(), int(new_n_samples));

        result = self.clone();
        if(mute_original):
            result.x = np.zeros(result.x.shape);
        if(gain_original is not None):
            result.x = result.x*gain_original;


        sl = len(s_toadd);
        for t in add_times:
            if(t<self.getDuration()):
                ti = int(t*self.sampling_rate);
                te = min(len(result.x), ti+sl);
                result.x[ti:te]=result.x[ti:te]+s_toadd[0:te-ti];
        result.setValueRange();
        # result.setMaxAbsValue(1.0);
        return result;


    def showSpectrogram(self, time_range = None, **kwargs):
        if(hop_length is None):
            hop_length=AUDIO_DEFAULT_HOP_LENGTH;

        fig = plt.figure();

        S = self.getSpectrogram(hop_length=hop_length, **kwargs);
        hop_length = self.getFeatureParams('spectrogram').get('hop_length');
        print(("hop length: {}".format(hop_length)));
        # S = librosa.stft(self.getSignal(), hop_length=hop_length);

        librosa.display.specshow(librosa.amplitude_to_db(S,ref = np.max),sr=self.sampling_rate, hop_length=hop_length, y_axis = 'linear', x_axis = 'time')
        plt.title('Power Spectrogram')
        plt.colorbar(format='%+2.0f dB')
        plt.ylim([0,8000]);
        plt.xlabel('Time (s)')
        if (time_range is not None):
            plt.xlim(time_range);
        plt.tight_layout()
        return fig;

    def showMelSpectrogram(self, force_recompute=False, **kwargs):
        mel_spec = self.getMelSpectrogram(force_recompute=force_recompute, **kwargs);
        # Make a new figure
        plt.figure();
        plt.figure(figsize=(12,4))
        # Display the spectrogram on a mel scale
        # sample rate and hop length parameters are used to render the time axis
        librosa.display.specshow(mel_spec, sr=self.sampling_rate, x_axis='time', y_axis='mel')
        # Put a descriptive title on the plot
        plt.title('Mel Power Spectrogram')
        # draw a color bar
        plt.colorbar(format='%+02.0f dB')
        # Make the figure layout compact
        plt.tight_layout()


    @staticmethod
    def Silence(n_seconds, sampling_rate, name=None):
        x = np.zeros(int(np.ceil(n_seconds*sampling_rate)));
        s = Audio(x=x, sampling_rate = sampling_rate, name = name);
        if(s.name is None):
            s.name = 'silence';
        return s;


    @staticmethod
    def _getDampedSin(freq, n_seconds=None, sampling_rate=16000, damping=0.1, noise_floor=0.001):
        """

        :param freq:
        :param n_seconds: if None, will set length necessary to damp to noise floor, up to 10.0s.
        :param sampling_rate:
        :param damping: Every period the amplitude is multiplied by 0.5^damping
        :param noise_floor:
        :return:
        """
        if(n_seconds is None):
            if(damping>0.005):
                nosc = truediv(math.log(noise_floor,0.5),damping);
                n_seconds = truediv(nosc, freq);
            else:
                n_seconds = 10.0;

        x = np.sin(np.linspace(0, n_seconds * freq * 2 * np.pi, np.round(n_seconds * sampling_rate)));
        dmp = np.linspace(0, n_seconds * freq, np.round(n_seconds * sampling_rate))
        dmp = np.power(0.5, dmp * damping);
        return np.multiply(x, dmp);



    @staticmethod
    def PingSound(n_seconds=None, freqs=None, damping=None, sampling_rate = 16000):
        if(freqs is None):
            # freqs = [400,500,600, 700, 800, 900, 1000, 1100, 1200, 1300];
            # freqs = np.arange(4, 25) * 100
            freqs = np.arange(5, 25) * 75; # just kind of thought this sounded fine...
        if(damping is None):
            damping = [0.05]*len(freqs);
        if(not isinstance(damping,list)):
            damping = [damping]*len(freqs);
        s = Audio._getDampedSin(freq=freqs[0], n_seconds = n_seconds, sampling_rate = sampling_rate, damping=damping[0]);
        for h in range(1,len(freqs)):
            new_s = Audio._getDampedSin(freq=freqs[h], n_seconds = n_seconds, sampling_rate = sampling_rate, damping=damping[h]);
            if(len(new_s)>len(s)):
                new_s[:len(s)]=new_s[:len(s)]+s;
                s = new_s;
            else:
                s[:len(new_s)] = s[:len(new_s)]+new_s;
        sa = Audio(x=s, sampling_rate=sampling_rate, name = 'ping');
        # sa.setValueRange(value_range=[-1,1]);
        sa.setMaxAbsValue(1.0);
        return sa;


    ##### --- FEATURES --- #####

    def getBeats(self, use_full_signal=True, tightness = None, force_recompute=False):
        """

        :param use_full_signal: If called from AudioClip class, this will determine whether the full signal is used or
            just the clip. This is important because the tempo is a global property. More evidence for a constant tempo
            is good, but if the tempo changes over the audio you probably don't want to use the full signal.
        :param tightness: How tightly to the tempo are beats picked. Must be positive. 0 would not care about tempo.
            100 is default. This is actually part of the penalty used in the dynamic programming objective from
            Ellis 2007. In librosa: txwt = -tightness * (np.log(-window / period) ** 2)
        :param force_recompute:
        :return:
        """
        if ((not self.hasFeature(name='beats')) or force_recompute):
            beat_args = dict(sr = self.sampling_rate, units = 'time');

            if (use_full_signal):
                beat_args.update(dict(y = self.getFullSignal()));
            else:
                beat_args.update(dict(y=self.getSignal()))
            if (tightness is not None):
                beat_args.update(dict(tightness=tightness));

            # print(beat_args)
            tempo, beats = librosa.beat.beat_track(**beat_args);
            self.setFeature(name='tempo', value=tempo);
            self.setFeature(name='beats', value=beats);
        return self.getFeature(name='beats');

    def getBeatVector(self, vector_length=None, force_recompute=False):
        """use_full_signal only makes a difference in AudioClip subclass."""
        if ((not self.hasFeature(name='beatvector')) or force_recompute):
            D = self.getDuration();
            if (vector_length is None):
                vector_length = int(math.ceil(D * 240));
            rvec = np.zeros(vector_length);
            beats = self.getFeature('beats');
            step = truediv(D, vector_length);
            for i in range(len(beats)):
                b = beats[i];
                id = int(truediv(b, step));
                rvec[id] = 1;
            self.setFeature(name='beatvector', value=rvec);
        return self.getFeature(name='beatvector');

    def getOnsets(self, use_full_signal=True, force_recompute=False, **kwargs):
        """use_full_signal only makes a difference in AudioClip subclass."""
        if ((not self.hasFeature(name='onsets')) or force_recompute):
            if (use_full_signal):
                onsets = librosa.onset.onset_detect(y=self.getFullSignal(), sr=self.sampling_rate, units='time',
                                                    **kwargs);
                self.setFeature(name='onsets', value=onsets);
            else:
                onsets = librosa.onset.onset_detect(y=self.getSignal(), sr=self.sampling_rate, units='time',
                                                    **kwargs);
                self.setFeature(name='onsets', value=onsets);
        return self.getFeature(name='onsets');

    def getOnsetSamplingRate(self):
        return np.true_divide(self.sampling_rate, AUDIO_DEFAULT_HOP_LENGTH);

    def pickOnsets(self, pre_max_time=0.03,
                   post_max_time=0.0,
                   pre_avg_time=0.1,
                   post_avg_time=0.1,
                   wait_time=0.03,
                   delta=0.07,
                   force_recompute=True, **kwargs):
        """

        :param pre_max_time:
        :param post_max_time:
        :param pre_avg_time:
        :param post_avg_time:
        :param wait_time:
        :param force_recompute:
        :param kwargs:
        :return:
        """
        # kwargs.setdefault('pre_max', 0.03 * sr // hop_length)  # 30ms
        # kwargs.setdefault('post_max', 0.00 * sr // hop_length + 1)  # 0ms
        # kwargs.setdefault('pre_avg', 0.10 * sr // hop_length)  # 100ms
        # kwargs.setdefault('post_avg', 0.10 * sr // hop_length + 1)  # 100ms
        # kwargs.setdefault('wait', 0.03 * sr // hop_length)  # 30ms
        # kwargs.setdefault('delta', 0.07)

        pick_params = dict(
            pre_max_time=pre_max_time,
            post_max_time=post_max_time,
            pre_avg_time=pre_avg_time,
            post_avg_time=post_avg_time,
            wait_time=wait_time,
            delta=delta,
        )

        tp_keys = list(pick_params.keys());
        for p in tp_keys:
            pick_params[p] = int(round(self.getOnsetSamplingRate() * pick_params[p]));

        dparams = dict(
            pre_max=pick_params['pre_max_time'],
            post_max=pick_params['post_max_time'] + 1,
            pre_avg=pick_params['pre_avg_time'],
            post_avg=pick_params['post_avg_time'] + 1,
            wait=pick_params['wait_time'],
            delta=delta
        )

        return self.getOnsets(force_recompute=force_recompute, **dparams);

    def getEvents(self):
        return self.getBeatEvents();

    def getEventList(self):
        return EventList(self.getBeatEvents());

    def getOnsetEvents(self):
        onsets = self.getOnsets();
        events = Event.FromStartTimes(onsets, type='onsets');
        return events;

    def getBeatEvents(self, start_time=None, end_time=None, **kwargs):
        beats = self.getBeats(**kwargs);
        start_beat = 0;
        end_beat = len(beats) - 1;
        if(start_time is not None):
            while((start_beat<len(beats)) and (beats[start_beat]<start_time)):
                start_beat=start_beat+1;
        if(end_time is not None):
            while(end_beat>0 and (beats[end_beat]>end_time)):
                end_beat = end_beat-1;
        if(end_beat>start_beat):
            return Event.FromStartTimes(beats[start_beat:end_beat], type='beats');
        else:
            return None;

    def getOnsetEnvelope(self, use_full_signal=True, force_recompute=False, centering=True, **kwargs):
        """use_full_signal only makes a difference in AudioClip subclass."""
        feature_name = 'onset_envelope';
        if((not self.hasFeature(name=feature_name)) or force_recompute):
            if(use_full_signal):
                eval_sig = self.getFullSignal();
            else:
                eval_sig = self.getSignal();
            onsets = librosa.onset.onset_strength(y=eval_sig, sr=self.sampling_rate, centering=centering, hop_length=AUDIO_DEFAULT_HOP_LENGTH, **kwargs);

            self.setFeature(name=feature_name, value=onsets);
        return self.getFeature(name=feature_name);


    def getMelSpectrogram(self, n_mels = 128, force_recompute=False):
        feature_name = 'melspectrogram';
        if((not self.hasFeature(name=feature_name)) or force_recompute):
            params = dict( sr = self.sampling_rate,
                                n_mels = n_mels);
            Spec = librosa.feature.melspectrogram(self.getSignal(), **params);
            self.setFeature(name=feature_name, value=librosa.power_to_db(Spec, ref=np.max), params=params);
        return self.getFeature(feature_name);

    def getSpectrogram(self, hop_length=None, force_recompute=False, **kwargs):
        feature_name = 'spectrogram';
        if((not self.hasFeature(name=feature_name)) or force_recompute):
            # params = dict( sr = self.sampling_rate);
            # params.update(kwargs);
            params = dict(kwargs);

            if(hop_length is None):
                hop_length = AUDIO_DEFAULT_HOP_LENGTH;
            params['hop_length'] = hop_length;

            center = kwargs.get('center');
            if(center is None):
                center = True;
            params['center'] = center;
            # print('recomputing {}'.format(hop_length))

            S = np.abs(librosa.stft(self.getSignal(), **params));
            self.setFeature(name=feature_name, value=S, params=params);
        return self.getFeature(feature_name);

    def getRMSE(self, force_recompute=False, hop_length=None, frame_length=None):
        feature_name = 'rmse';
        if((not self.hasFeature(name=feature_name)) or force_recompute):
            if(frame_length is None):
                frac_of_second = 0.05;
                frame_length=max(1, int(self.sampling_rate*frac_of_second));
            if(hop_length is None):
                hop_length=int(math.floor(frame_length*0.5));
            params = dict( hop_length = hop_length,
                                center = True,
                                frame_length = frame_length);
            rmse = librosa.feature.rmse(y=self.getSignal(), **params);
            self.setFeature(name=feature_name, value=np.ndarray.flatten(rmse), params=params);
        return self.getFeature(feature_name);


    def getBeatTimeBefore(self, t):
        return self.getBeatBefore(t=t).start;

    def getBeatBefore(self, t):
        beats = self.getBeatEvents();
        bi = self.getBeatIndexBefore(t=t);
        return beats[bi].start;

    def getBeatIndexBefore(self, t):
        beats = self.getBeatEvents();
        for i, b in enumerate(beats):
            if(b.start>t):
                return i-1;
        return len(beats)-1;


    def getTempogram(self, window_length=None, force_recompute=None, frame_rate=None, resample_rate = None, **kwargs):
        """

        :param self:
        :param window_length: in seconds
        :param force_recompute:
        :param kwargs:
        :return:
        """
        feature_name = 'tempogram';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            if (window_length is None):
                window_length = DEFAULT_TEMPOGRAM_WINDOW_SECONDS;
            tpgparams = {};
            tpgparams.update(kwargs);

            sr = self.sampling_rate;
            y=self.getSignal();
            if(resample_rate is None):
                resample_rate = 22050;
            if(sr>resample_rate):
                print(("resampling {}Hz to {}Hz".format(sr, resample_rate)));
                y = librosa.core.resample(y, orig_sr=sr, target_sr=resample_rate);
                sr = resample_rate;
                print("resampled")

            if(frame_rate is None):
                frame_rate = 30;
            hop_length = int(round(truediv(sr,frame_rate)));
            win_length = int(round(window_length * frame_rate));

            tparams = dict(y=y,
                           sr=sr,
                           hop_length=hop_length,
                           win_length=win_length,
                           **kwargs);
            tpgparams.update(dict(sr=sr,hop_length=hop_length, win_length=win_length));
            result = librosa.feature.tempogram(**tparams);
            ###########
            tempo_bpms = librosa.tempo_frequencies(result.shape[0], hop_length=hop_length, sr=sr)
            self.setFeature(name='tempogram_bpms', value=tempo_bpms);
            self.setFeature(name=feature_name, value=result, params=tpgparams);
            self.setInfo(label='tempogram_params',value=tpgparams);

        return self.getFeature(feature_name);

    def plotTempogram(self, window=None, time_range=None, **kwargs):
        tempogram = self.getFeature('tempogram', force_recompute=True);
        tparams = self.getInfo('tempogram_params')
        toshow = tempogram;
        if(window is not None):
            wstart=int(round(window[0]*self.sampling_rate));
            wend = int(round(window[1]*self.sampling_rate));
            toshow = tempogram[:,wstart:wend];

        mplt = librosa.display.specshow(toshow, sr=tparams['sr'], hop_length=tparams['hop_length'], x_axis = 'time', y_axis = 'tempo')
        plt.legend(frameon=True, framealpha=0.75)
        plt.set_cmap('coolwarm')
        plt.colorbar(format='%+2.0f dB')
        if (time_range is not None):
            plt.xlim(time_range);
        plt.xlabel('Time (s)')
        plt.title('Audio Tempogram');
        plt.tight_layout()
        return mplt;

    def plotOnsets(self, **kwargs):
        signal = self.getFeature('onset_envelope');
        events = self.getOnsetEvents();
        mplt = Event.PlotSignalAndEvents(signal, sampling_rate=self.getOnsetSamplingRate(), events=events, **kwargs);
        plt.xlabel('Time (s)')
        plt.ylabel('Onset Strength')
        return mplt;

    def plotBeats(self, **kwargs):
        signal = self.getFeature('onset_envelope');
        events = self.getBeatEvents();
        mplt = Event.PlotSignalAndEvents(signal, sampling_rate=self.getOnsetSamplingRate(), events=events, **kwargs);
        plt.title('Onset Envelope and Beats');
        plt.xlabel('Time (s)')
        plt.ylabel('Onset Strength')
        return mplt;

    def plotOnsetEnvelope(self, **kwargs):
        signal = self.getFeature('onset_envelope');
        events = None;
        mplt = Event.PlotSignalAndEvents(signal, sampling_rate=self.getOnsetSamplingRate(), events=events, **kwargs);
        plt.title('Onset Envelope');
        plt.xlabel('Time (s)')
        plt.ylabel('Onset Strength')
        return mplt;

    def plotSignal(self, time_range=None, ylim=None, **kwargs):
        signal = self.getSignal();
        # Event.PlotSignalAndEvents(signal, sampling_rate=self.getOnsetSamplingRate(), events=events, **kwargs);
        times = np.arange(len(signal));
        times = times * truediv(1.0, self.sampling_rate);
        mplt = plt.plot(times, signal);
        if (time_range is not None):
            plt.xlim(time_range[0], time_range[1])
        if (ylim is not None):
            plt.ylim(ylim);
        plt.title('Time Signal');
        return mplt;


    FEATURE_FUNCS['melspectrogram'] = getMelSpectrogram;
    FEATURE_FUNCS['spectrogram'] = getSpectrogram;
    FEATURE_FUNCS['rmse'] = getRMSE;
    FEATURE_FUNCS['beats'] = getBeats;
    FEATURE_FUNCS['onsets'] = getOnsets;
    FEATURE_FUNCS['beatvector'] = getBeatVector;
    FEATURE_FUNCS['tempogram'] = getTempogram;
    FEATURE_FUNCS['onset_envelope'] = getOnsetEnvelope;


def _make_wav(data, rate):
    """ Transform a numpy array to a PCM bytestring """
    import struct
    from io import BytesIO
    import wave

    try:
        import numpy as np

        data = np.array(data, dtype=float)
        if len(data.shape) == 1:
            nchan = 1
        elif len(data.shape) == 2:
            # In wave files,channels are interleaved. E.g.,
            # "L1R1L2R2..." for stereo. See
            # http://msdn.microsoft.com/en-us/library/windows/hardware/dn653308(v=vs.85).aspx
            # for channel ordering
            nchan = data.shape[0]
            data = data.T.ravel()
        else:
            raise ValueError('Array audio input must be a 1D or 2D array')
        scaled = np.int16(data / np.max(np.abs(data)) * 32767).tolist()
    except ImportError:
        # check that it is a "1D" list
        idata = iter(data)  # fails if not an iterable
        try:
            iter(next(idata))
            raise TypeError('Only lists of mono audio are '
                            'supported if numpy is not installed')
        except TypeError:
            # this means it's not a nested list, which is what we want
            pass
        maxabsvalue = float(max([abs(x) for x in data]))
        scaled = [int(x / maxabsvalue * 32767) for x in data]
        nchan = 1

    fp = BytesIO()
    waveobj = wave.open(fp, mode='wb')
    waveobj.setnchannels(nchan)
    waveobj.setframerate(rate)
    waveobj.setsampwidth(2)
    waveobj.setcomptype('NONE', 'NONE')
    waveobj.writeframes(b''.join([struct.pack('<h', x) for x in scaled]))
    val = fp.getvalue()
    waveobj.close()

    return val


from visbeat.AImports import *

VIEWER_INSTALLED = 1;
try:
    import vbwidget as Viewer
except ImportError as e:
    VIEWER_INSTALLED = 0;
    AWARN("VBViewer not installed. Consider installing for full functionality.")

from ..TimeSignal import *
from ..EventList import *



#this is what media should call to get its gui object
def media_GUI_func(self):
    if (self._gui is None):
        self._gui = BeatGUI();
        self._gui.media = self;
    return self._gui;

class BeatGUI(AObject):
    """

    """
    def AOBJECT_TYPE(self):
        return 'BeatGUI';

    def __init__(self, media=None, path=None, clear_temp=None):
        """If you provide a directory, it will look for a existing AFileManager.json in that directory, or create one if it does not already exist.
            If you provide a json, it will use that json, unless the json doesn't exist, in which case it will complain...
        """
        AObject.__init__(self, path=path);
        if(media is not None):
            self.media = media;

    def initializeBlank(self):
        AObject.initializeBlank(self);
        self._widget = None;
        self._media = None;


    def getJSONName(self):
        return self.AOBJECT_TYPE()+".json";


    #########

    # <editor-fold desc="Property: 'media'">
    @property
    def media(self):
        return self._getMedia();
    def _getMedia(self):
        return self._media;
    @media.setter
    def media(self, value):
        self._setMedia(value);
    def _setMedia(self, value):
        self._media = value;
    # </editor-fold>

    # <editor-fold desc="Property: 'media_type'">
    @property
    def media_type(self):
        return self._getMediaType();
    def _getMediaType(self):
        if(self.media is None):
            return None;
        else:
            return self.media.AObjectType();
    # </editor-fold>

    # <editor-fold desc="Property: 'widget'">
    @property
    def widget(self):
        return self._getWidget();
    def _getWidget(self):
        if (self._widget is None):
            self._widget = Viewer.VBVSignal();
        return self._widget;
    @widget.setter
    def widget(self, value):
        self._setWidget(value);
    def _setWidget(self, value):
        self._widget = value;
    # </editor-fold>

    # <editor-fold desc="Property: 'frame_rate'">
    @property
    def frame_rate(self):
        return self._getFrameRate();
    def _getFrameRate(self):
        gfr = self.widget.frame_rate;
        if (gfr is None):
            media = self.media;
            if (media is not None):
                gfr = media.getFrameRate();
        return gfr;
    @frame_rate.setter
    def frame_rate(self, value):
        self._setFrameRate(value);
    def _setFrameRate(self, value):
        self.widget.frame_rate = float(value);
    # </editor-fold>


    # <editor-fold desc="Property: 'frame_offset'">
    @property
    def frame_offset(self):
        return self._getFrameOffset();
    def _getFrameOffset(self):
        return self.widget.frame_offset;
    @frame_offset.setter
    def frame_offset(self, value):
        self._setFrameOffset(value);
    def _setFrameOffset(self, value):
        self.widget.frame_offset = value;
    # </editor-fold>

    def run(self, local_saliency=None, frame_rate = None, eventlist = 'default', frame_offset=None):
        if(frame_rate is None):
            # self.widget.frame_rate = float(self.getMedia().getFrameRate());
            self.frame_rate = self.media._getFrameRate();
        else:
            # self.widget.frame_rate = float(frame_rate);
            self.frame_rate = frame_rate;

        if(local_saliency is None):
            self.widget.signal = self.media.getLocalRhythmicSaliency().tolist();
            # self.widget.signal = self.getBothWayVisualImpactEnvelope(highpass_window_seconds=None, force_recompute = True).tolist();
        else:
            self.widget.signal = local_saliency.tolist();

        if(frame_offset is None):
            self.frame_offset = 0;
        elif(frame_offset is 'guess'):
            self.frame_offset = self.guessFrameOffset();
        else:
            self.frame_offset = frame_offset;

        if(eventlist is None):
            self.widget.events = [];
        elif(eventlist == 'default'):
            self.widget.events = EventList._toGUIDicts(self.media.getEventList());
        else:
            self.widget.events = EventList._toGUIDicts(eventlist);
        self.widget.data_string = self.media.getStringForHTMLStreamingBase64();
        return self.widget;


    def guessFrameOffset(self):
        if(isinstance(self.media, Video)):
            return self.media.reader.get_length() - self.media.n_frames();
        else:
            return 0;


    def deactivateAllEvents(self):
        newes = []
        gevents = self.getEventDicts();
        for e in gevents:
            newe = e;
            newe['is_active']=0;
            newes.append(newe);
        self.widget.events = [];
        self.widget.events = newes;

    def activateAllEvents(self):
        newes = []
        gevents = self.getEventDicts();
        for e in gevents:
            newe = e;
            newe['is_active'] = 1;
            newes.append(newe);
        self.widget.events = [];
        self.widget.events = newes;

    def activatePattern(self, pattern=None, prefix=None, apply_to_active=None):
        assert(pattern), "must provide pattern to activate"
        newes = []
        gevents = self.getGUIEventDicts();
        counter = 0;
        prefix_length = 0;
        if(prefix_length is not None):
            prefix_length = len(prefix);
        for i, e in enumerate(gevents):
            if (apply_to_active):
                if (e.get('is_active')):
                    if (counter < prefix_length):
                        e['is_active']=prefix[counter];
                    else:
                        e['is_active'] = pattern[(counter - prefix_length) % len(pattern)];
                    counter = counter + 1;
                else:
                    print(("Skipping beat {}, inactive".format(i)));
            else:
                if (i < prefix_length):
                    e['is_active'] = prefix[i];
                else:
                    e['is_active'] = pattern[(i - prefix_length) % len(pattern)];
            newes.append(e);

        self.widget.events = [];
        self.widget.events = newes;


    def shiftEventsByNFrames(self, n_frames=None):
        assert(n_frames), "must provide number of frames to shift by"
        newes = []
        gevents = self.getEventDicts();
        sample_step = np.true_divide(1.0,self.getFrameRate());
        for e in gevents:
            newe = e;
            newe['start'] = newe['start']+n_frames*sample_step;
            newes.append(newe);
        self.widget.events = [];
        self.widget.events = newes;


    def getActiveEventTimes(self):
        gevents = self.getEventDicts(active_only=True);
        revents = []
        for e in gevents:
            revents.append(e.get('time'));
        return np.asarray(revents);

    def getEventTimes(self):
        gevents = self.getEventDicts();
        revents = []
        for e in gevents:
            revents.append(e.t);
        return np.asarray(revents);

    def getEvents(self, active_only=None):
        return Event._FromGUIDicts(self.getEventDicts(active_only = active_only));

    def getEventList(self, active_only=None):
        elist = EventList._FromGUIDicts(self.getEventDicts(active_only=active_only));
        elist.setInfo(label='html_frame_offset', value=self.getFrameOffset());
        return elist;

    def getActiveEvents(self):
        return self.getEvents(active_only=True);

    def getEventDicts(self, active_only = None):
        gevents = self.widget.events[:];
        if(not active_only):
            return gevents;
        else:
            nevents = []
            for e in gevents:
                if(e.get('is_active')):
                    nevents.append(e);
            return nevents;

    def saveEvents(self, save_path = None):
        elist = self.getEventList(active_only=False);
        if(save_path is not None):
            elist.writeToJSON(json_path=save_path);
            self.widget.last_save_path = save_path;
        else:
            save_path = self.widget.last_save_path;
            if(save_path is None):
                save_path = uiGetSaveFilePath(file_extension='.json');
            if(save_path is not None):
                elist.writeToJSON(json_path=save_path);
                self.widget.last_save_path = save_path;

    def saveEventsAs(self, save_path = None):
        elist = self.getEventList(active_only=False);
        if (save_path is not None):
            elist.writeToJSON(json_path=save_path);
            self.widget.last_save_path = save_path;
            print(('savepath not none {}'.format(save_path)))
        else:
            save_path = uiGetSaveFilePath(file_extension='.json');
            print(('savepath from ui {}'.format(save_path)))
            if (save_path is not None):
                print(('save path from ui {}'.format(save_path)));
                elist.writeToJSON(json_path=save_path);
                self.widget.last_save_path = save_path;
        print(save_path)

    def setEvents(self, events):
        self.widget.events = Event._ToGUIDicts(events);

    def setEventList(self, event_list):
        if(event_list.getInfo('html_frame_offset') is not None):
            self.widget.frame_offset = event_list.getInfo('html_frame_offset');
        self.widget.events = event_list._toGUIDicts();


    def loadEvents(self, load_path = None):
        if(load_path is None):
            load_path = uiGetFilePath();
        elist = EventList();
        elist.loadFromJSON(json_path=load_path);
        self.setEventList(event_list = elist);



    def getEventListWithSelectedSegments(self):
        eventlist = self.getEventList();
        events = eventlist.events;
        segments = [];
        for i, e in enumerate(events):
            if(e.direction>-1): # meaning not a back beat
                newseg = [];
                for si in range(i, len(events)):
                    newseg.append(si);
                    if(events[si].direction<0): # meaning a back beat
                        break;
                segments.append(newseg);
        eventlist.setInfo(label='selected_segments', value=segments);
        return eventlist;








import os
import subprocess



def uiGetFilePath(initial_path=None):
    try:
        if(initial_path):
            output = subprocess.check_output("osascript -e 'set strPath to POSIX file \"{}\"' -e 'set theDocument to choose file with prompt \"Please select a document to process:\" default location strPath' -e 'set theDocument to (the POSIX path of theDocument)'".format(initial_path), shell=True)
        else:
            output = subprocess.check_output("osascript -e 'set theDocument to choose file with prompt \"Please select a document to process:\"' -e 'set theDocument to (the POSIX path of theDocument)'", shell=True)
        return output.replace('\n', '');
    except subprocess.CalledProcessError as e:
        print((e.output));
    # assert(False)
    #     grabpath = get_ipython().run_cell_magic(u'bash', u'', "osascript -e 'set theDocument to choose file with prompt \"Please select a document to process:\"' -e 'set theDocument to (the POSIX path of theDocument)'>&2")

def uiGetDirectory(initial_path=None):
    try:
        if(initial_path):
            output = subprocess.check_output("osascript -e 'set strPath to POSIX file \"{}\"' -e 'set thedir to choose folder with prompt \"Please select a file:\" default location strPath' -e 'set thedir to (the POSIX path of thedir)'".format(initial_path), shell=True)
        else:
            output = subprocess.check_output("osascript -e 'set thedir to choose folder with prompt \"Please select a directory:\"' -e 'set thedir to (the POSIX path of thedir)'", shell=True)
        return output.replace('\n', '');
    except subprocess.CalledProcessError as e:
        print((e.output));
    # assert(False)

def uiGetSaveFilePath(initial_path=None, file_extension=None):
    try:
        osastr = "osascript ";
        if(initial_path):
            osastr = osastr+"-e 'set strPath to POSIX file \"{}\"' ".format(initial_path);
        osastr = osastr+"-e 'set theDocument to choose file name with prompt \"Save As File:\" ";
        if(initial_path):
            osastr = osastr+"default location strPath";
        osastr = osastr+"' ";
        osastr = osastr+"-e 'set theDocument to (the POSIX path of theDocument)'"
        output = subprocess.check_output(osastr, shell=True);
        ostring = output.replace('\n', '');
        if (file_extension is not None):
            if (not ostring.endswith(file_extension)):
                ostring = ostring + file_extension;
        return ostring;
    except subprocess.CalledProcessError as e:
        AWARN('ERROR')
        print((e.output));

def showInFinder(path):
    return openOSX(get_dir_from_path(path));

def openOSX(path):
    return subprocess.check_output("open {}".format(put_string_in_quotes(path)), shell=True);

def put_string_in_quotes(s):
    return "\""+s+"\""

def get_file_name_from_path(pth):
    return os.path.split(pth)[1];

def get_dir_from_path(pth):
    return (os.path.split(pth)[0]+os.sep);


from .AParamDict import *
import pickle as pickle
import os
class AFuncDict(AParamDict):
    """AFuncDict (class): Extends AParamDict so that functions can be assigned to features and called whenever
        computing those features is necessary.
        Attributes:
            data: name -> value, params
            feature funcs: name -> function for evaluating
    """

    def __init__(self, owner=None, name=None, path=None):
        AParamDict.__init__(self, owner=owner, name=name, path=path);
        self.functions = {};

    def getEntry(self, name=None, params=None, force_recompute=False):
        d = self.data.get(name);
        if((d is not None) and (not force_recompute)):
            return d;
        else:
            f = self.getFunction(name=name);
            if(f is not None):
                if(params is not None):
                    if(not params.get('force_recompute')):
                        params.update(dict(force_recompute=force_recompute));
                    self.setValue(name=name, value=f(self=self.owner, **params), params=params, modified=True);
                else:
                    self.setValue(name=name, value=f(self=self.owner, force_recompute=force_recompute), params=params, modified=True);
            return self.data.get(name);
        return None;

    def getValue(self, name=None, params=None, force_recompute=False):
        d = self.getEntry(name=name, params=params, force_recompute=force_recompute);
        if(d is not None):
            return d.get('value');
        else:
            return None;

    def getParams(self, name=None):
        d = self.data.get(name);
        if(d is not None):
            return d.get('params');
        else:
            return None;

    def getFunction(self, name=None):
        return self.functions.get(name);

    def setValue(self, name, value=None, params=None, modified=True):
        self.data[name]['value']=value;
        self.data[name]['params']=params;
        self.setEntryModified(name=name, is_modified=modified)
        #self.data[name]['modified']=modified;

    def setFunction(self, name, function=None):
        self.functions[name]=function;

    def saveEntry(self, name, path, force=False):
        """Save one entry to one file."""
        if(self.data.get(name) is None):
            return None;
        if(self.isEntryModified(name=name) or force or (not os.path.isfile(path))):
            #pickleToPath
            f = open(path, 'wb');
            pickle.dump(self.getEntry(name=name), f, protocol=2);
            f.close();
            self.setEntryModified(name=name, is_modified=False);
            #assert(False), "should not be saving in this test";
        return True;

    def setEntryModified(self, name, is_modified=True):
        self.data[name]['modified']=is_modified;
        if(is_modified):
            self.setModified(is_modified=True);

    def isEntryModified(self, name):
        entry = self.data.get(name);
        if(entry is not None):
            m=entry.get('modified');
            if(m is not None):
                return m;
            else:
                return True;
        else:
            assert(False), "checking mod bit on entry that does not exist"

    def isModified(self):
        return self.modified;

    def setModified(self, is_modified):
        self.modified=is_modified;

    def save(self, path, force=False):
        """save all entries to one file."""
        if(force or self.isModified()):
            f = open(path, 'wb');
            pickle.dump(self.data, f, protocol=2);
            f.close();
            self.setModified(is_modified=False);
            #assert(False), "should not be saving in this test";
        return True;

    def loadEntry(self, name, path):
        """load one entry from one file."""
        f=open(path, 'rb');
        self.setEntry(name=name, d=pickle.load(f));
        f.close();
        self.setEntryModified(name=name, is_modified=False);
        return True;

    def load(self, path):
        """Load a set of entries all from one file."""
        f=open(path, 'rb');
        newd = pickle.load(f);
        self.data.update(newd);
        f.close();
        return True;

    def getKeyList(self):
        return list(self.data.keys());

    def getFunctionList(self):
        return list(self.functions.keys());



from .Audio import *

class AudioClip(Audio):
    """AudioClip (class): A segment of a video, and a bunch of convenience functions to go with it.
        Attributes:
            start: The name of the video
            end: The framerate of the video
    """

    def VBOBJECT_TYPE(self):
        return 'AudioClip';

    def __init__(self, audio=None, start=None, end=None, path=None):
        assert(not (audio and path)), "provided both video object and path to VideoClip constructor."
        # assert((start is not None) and (end is not None)), "must provide start time and end time to AudioClip constructor"
        Audio.__init__(self, path=path);
        #self.initializeBlank();#gets called by parent
        if(start is None):
            start = 0;
        if(end is None):
            if(audio is not None):
                end = audio.getDuration();
            else:
                end = truediv(len(self.x), self.sampling_rate);

        self.start = start;
        self.end = end;


        if(audio):
            self.setPath(audio.getPath());
            self.x = audio.x;
            self.sampling_rate=audio.sampling_rate;
            self.n_channels=audio.n_channels;
            stereo = audio.getStereo();
            if(stereo):
                self.a_info['stereo_signal']=stereo;
        if(self.sampling_rate):
            self._pull_clip_potion();


    def _pull_clip_potion(self):
        startsample = math.floor(self.sampling_rate * self.start);
        endsample = math.ceil(self.sampling_rate * self.end);
        startinsample = max(startsample, 0);
        endinsample = min(endsample, len(self.x));
        self.clipped = self.x[int(startinsample):int(endinsample)];
        if (startsample < 0):
            self.clipped = np.concatenate((np.zeros(int(-startsample)), self.clipped));
        if (endsample > len(self.x)):
            self.clipped = np.concatenate((np.zeros(int(endsample - len(self.x))), self.clipped));


    def resample(self, sampling_rate):
        Audio.resample(self, sampling_rate);
        self._pull_clip_potion();

    def initializeBlank(self):
        Audio.initializeBlank(self);
        self.start = None;
        self.end = None;
        self.resampled = None;
        self.clipped = None;

    def getSignal(self, resample=False):
        if(self.resampled):
            return self.resampled;
        if(resample):
            assert(False), "haven't implemented audio clip resampling yet.";
        return self.clipped



#AO#labels#AObject
import os
import json

from .AParamDict import *
from . import fileui

class AObject(object):
    """AObject (class): This is a paarent class used to implement any comon serialization or typing we might want to do later
        Attributes:
            labels: dictionary of meta_data
            save, load, and clear funcs: these are hooks for functions that manage the object's data on disk.
    """

    AOBJECT_BASE_PATH=None;

    def __init__(self, path=None, **kwargs):
        self.initializeBlank();
        if(path):
            self.setPath(file_path=path, **kwargs);


    def initializeBlank(self):
        self.a_info = {'AObjectType': self.AOBJECT_TYPE()};
        # self.a_data = AParamDict(owner=self, name='a_data');
        self.save_func = None;
        self.load_func = None;
        self.clear_func = None;

    def setPath(self, file_path=None, **kwargs):
        if(file_path):
            self.a_info['file_path'] = pathstring(file_path);
            pparts = os.path.split(self.a_info['file_path']);
            self.a_info['file_name'] = pparts[1];
            self.a_info['directory_path']=pparts[0];
            filename = self.a_info.get('file_name');
            if(filename):
                nameparts = os.path.splitext(filename);
                self.a_info['file_base_name'] = nameparts[0];
                self.a_info['file_ext'] = nameparts[1];
            self.a_info['base_path'] = kwargs.get('base_path');
            if(self.a_info['base_path'] is None):
                self.a_info['base_path'] = AObject.AOBJECT_BASE_PATH;

    def getPath(self):
        if('file_path' in self.a_info):
            return self.a_info['file_path'];
        else:
            return None;

    def _showFile(self):
        if(fileui.Show is not None):
            fileui.Show(self.getPath());

    def _open(self):
        if(fileui.Open is not None):
            fileui.Open(self.getPath());

    def getRelativePath(self, base_path=None):
        base = base_path;
        if(base is None):
            base = self.a_info.get('base_path');
        if(base is None):
            AWARN("Base path not set!");
        return str(self.getPath());

    def getFileName(self):
        if('file_name' in self.a_info):
            return self.a_info['file_name'];
        else:
            return None;

    def getFileExtension(self):
        if('file_ext' in self.a_info):
            return self.a_info['file_ext'];
        else:
            return None;

    def getDirectoryPath(self):
        return self.a_info.get('directory_path')

    def setInfo(self, label, value):
        self.a_info[label]=value;

    def getInfo(self, label):
        return self.a_info.get(label);

    def loadFromJSON(self, json_path=None):
        if(json_path):
            self.setPath(file_path=json_path);
        if('file_path' in self.a_info):
            json_text=open(self.a_info['file_path']).read();
            d = json.loads(json_text);
            self.initFromDictionary(d);

    def writeToJSON(self, json_path=None):
        #with open(jsonpath+self.name+'.json', 'w') as outfile:
        if(not json_path):
            json_path = self.a_info.get('file_path');
        if(json_path):
            with open(json_path, 'w') as outfile:
                json.dump(self.toDictionary(), outfile, sort_keys = True, indent = 4, ensure_ascii=False);

    def serializeInfo(self):
        return self.a_info;

    def save(self, features_to_save='all', overwrite=True, **kwargs):
        if(self.save_func):
            self.save_func(self, features_to_save=features_to_save, overwrite=overwrite, **kwargs);
        else:
            AWARN("SAVE FUNCTION HAS NOT BEEN PROVIDED FOR {} INSTANCE".format(self.AOBJECT_TYPE()));

    def load(self, features_to_load=None, **kwargs):
        if(self.load_func):
            self.load_func(self, features_to_load=features_to_load, **kwargs);
        else:
            AWARN("LOAD FUNCTION HAS NOT BEEN PROVIDED FOR {} INSTANCE".format(self.AOBJECT_TYPE()));

    ########### "VIRTUAL" FUNCTIONS #############
    @staticmethod
    def AOBJECT_TYPE():
        return 'AObject';

    def toDictionary(self):
        d = {'a_info': self.serializeInfo()};
        return d;

    def initFromDictionary(self, d):
        self.a_info = d.get('a_info');

    # ##Example of how these functions should be written in a  subclass, here we call the subclass 'AssetManager'
    # def AOBJECT_TYPE(self):
    #     return 'AssetManager';
    #
    # def toDictionary(self):
    #     d = AObject.toDictionary(self);
    #     #serialize class specific members
    #     return d;
    #
    # def initFromDictionary(self, d):
    #     AObject.initFromDictionary(self, d);
    #     #do class specific inits with d;





from .Event import *

class EventList(AObject):
    """Event (class): An event in time, either in video or audio
        Attributes:
            start: when the event starts
    """

    def AOBJECT_TYPE(self):
        return 'EventList';

    def __init__(self, events=None):
        AObject.__init__(self, path=None);
        if(events is not None):
            self.events = events;


    def initializeBlank(self):
        AObject.initializeBlank(self);
        self.events = [];

    def list(self):
        return self.events;

    def Clone(self):
        return EventList(Event.Clone(self.events));

    def toDictionary(self):
        d=AObject.toDictionary(self);
        d['events']=self.serializeEvents();
        return d;

    def serializeEvents(self):
        re = [];
        for e in self.events:
            re.append(e.toDictionary());
        return re;

    def getActiveEvents(self):
        active_events = [];
        for e in self.events:
            if(e.is_active):
                active_events.append(e);
        return active_events;

    def initFromDictionary(self, d):
        AObject.initFromDictionary(self, d);
        events = d['events'];
        for e in events:
            newe = Event();
            newe.initFromDictionary(e);
            self.events.append(newe);

    def unroll(self, assert_on_folds=None):
        newlist = Event.GetUnrolledList(event_list=self.events, assert_on_folds=assert_on_folds);
        self.events = newlist;

    def getUnrolled(self, assert_on_folds=None):
        return EventList(Event.GetUnrolledList(event_list=self.events, assert_on_folds=assert_on_folds));

    def getFromIndices(self, inds):
        return EventList(Event.NewFromIndices(event_list=self.events, inds=inds));

    def getRolledToN(self, n_out, momentum = 0.1):
        return EventList(Event.RollToN(self.events, n_out=n_out, momentum=momentum));


    def toStartTimes(self):
        return Event.ToStartTimes(self.events);

    def _toGUIDicts(self):
        return Event._ToGUIDicts(self.events);

    @staticmethod
    def _FromGUIDicts(gds, type=None):
        events = Event._FromGUIDicts(gds, type=type);
        return EventList(events);


    @staticmethod
    def FromJSON(json_path=None):
        if (json_path is None):
            json_path = fileui.GetFilePath();
        elist = EventList();
        elist.loadFromJSON(json_path=json_path);
        return elist;

    @staticmethod
    def FromStartTimes(starts, type=None, event_class=None):
        elist = EventList();
        if(event_class is None):
            event_class = Event;
        for s in starts:
            elist.events.append(event_class(start=s, type=type));
        return elist;


    def toWeights(self):
        return Event.ToWeights(self.events);

    def getDoubled(self):
        return EventList(Event.Double(self.Clone().events));

    def getHalved(self, offset=0):
        return EventList(Event.Half(self.Clone().events, offset=offset));

    def getThirded(self, offset=0):
        return EventList(Event.Third(self.Clone().events, offset=offset));

    @staticmethod
    def FromSignalPeaks(**kwargs):
        return EventList(Event.FromSignalPeaks(**kwargs));

    @staticmethod
    def PlotEventMatches(source_elist, target_elist, source_in=None, target_in=None):
        return Event.PlotEventMatches(source_events=source_elist.events, target_events=target_elist.events, source_in=source_in, target_in=target_in);

    def getWithFirstEventAt(self, first_event_time):
        return EventList(Event.GetWithFirstEventAt(events=self.events, first_event_time=first_event_time));

    def getWithStartTimesShifted(self, new_start_time):
        return EventList(Event.GetWithStartTimesShifted(events = self.events, new_start_time=new_start_time));


# from Video import *
from .Image import *
from .Event import *
from .VisualBeat import *
import librosa

FEATURE_FUNCS = {};
VIS_FUNCS = {};
VISVIDEO_FUNCS = {};

FLOW_LOG_EPSILON=1.0;
FLOW_UNIT_GAIN=10000.0;

HISTOGRAM_FRAMES_PER_BEAT = 2;
HISTOGRAM_DOWNSAMPLE_LEVELS = 3;


VB_UPSAMPLE_FACTOR = 1.0;
USING_OPENCV = Image.USING_OPENCV;

if(USING_OPENCV):

    ##########################################################################
    # ################ THESE ARE THE FUNCTIONS TO OVERRIDE! ################ #
    # ################     FOR CUSTOM SALIENCY METRICS!     ################ #
    ##########################################################################

    # You can modify them here. But I would suggest elsewhere in your code
    # just setting Video.localRhythmicSaliencyFunction and
    # Video.visualBeatFunction to whatever you want. See how they are assigned
    # when Video_CV is included in Video.py. -Abe

    def localRhythmicSaliencyFunction(self, **kwargs):
        """
        Change to use different function for local saliency
        """
        return self.getVisibleImpactEnvelope(**kwargs);

    def visualBeatFunction(self, **kwargs):
        """
        Change to use different default strategy for selecting visual beats
        """
        # beat_params = dict(
        #     pre_max_time=0.2,
        #     post_max_time=0.2,
        #     pre_avg_time=0.2,
        #     post_avg_time=0.2,
        #     wait_time=0.1,
        # )
        beat_params = self._getDefaultPeakPickingTimeParams();
        beat_params.update(kwargs);
        return self.getVisibleImpacts(**beat_params);

    # #################### This is how they are called #################### #

    def getLocalRhythmicSaliency(self, force_recompute=False, **kwargs):
        feature_name = 'local_rhythmic_saliency';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = dict();
            params.update(kwargs);
            params.update({'force_recompute':force_recompute});
            result = self.localRhythmicSaliencyFunction(**params);
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    def getVisualBeats(self, force_recompute=False, **kwargs):
        feature_name = 'visual_beats';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            params.update({'force_recompute': force_recompute});
            result = self.visualBeatFunction(**params);
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    # ##################################################################### #

    ##########################################################################
    # ###################################################################### #
    ##########################################################################



    def cvGetGrayFrame(self, f):
        colorframe = self.getFrame(f).astype(np.uint8);
        return ocv.cvtColor(colorframe, ocv.COLOR_RGB2GRAY);

    def getImageFromFrameGray(self, f):
        frame = self.getImageFromFrame(f);
        frame.RGB2Gray();
        return frame;

    def flow2row(ang, amp, bins, subdivs, n_shifts, density):
        h, w = ang.shape[:2];
        ncells = np.power(4, subdivs);
        nperd = np.power(2, subdivs);

        xw = w-n_shifts[1];
        yw = h-n_shifts[0];

        xcells = np.arange(nperd + 1, dtype=float);
        ycells = np.arange(nperd + 1, dtype=float);
        xcells = np.floor((xcells / (nperd)) * xw);
        ycells = np.floor((ycells / (nperd)) * yw);
        # print(n_shifts)

        ahis = np.zeros([ncells * bins, n_shifts[0]*n_shifts[1]]);

        for dy in range(n_shifts[0]):
            for dx in range(n_shifts[1]):
                ampwin = amp[dy:dy+yw,dx:dx+xw];
                angwin = ang[dy:dy+yw,dx:dx+xw];
                cell_counter = 0;
                for x in range(nperd):
                    for y in range(nperd):
                        ystart=int(ycells[y]);
                        yend = int(ycells[y+1]);
                        xstart = int(xcells[x]);
                        xend = int(xcells[x + 1]);
                        angcell = angwin[ystart:yend, xstart:xend];
                        ampcell = ampwin[ystart:yend, xstart:xend];
                        cahis, cbinbounds = np.histogram(angcell.ravel(), bins=bins, range=(0, 2 * np.pi),
                                                         weights=ampcell.ravel(), density=density);
                        # print("ahis shape: {}\ncahis shape: {}\ndx, dy: {}, {}\ncell_counter: {}\nbins: {}\n".format(ahis.shape, cahis.shape, dx, dy, cell_counter, bins));
                        ahis[cell_counter * bins:(cell_counter + 1) * bins, dx+dy*n_shifts[0]] = cahis;
                        cell_counter = cell_counter + 1;
        return ahis;

    def getFlowFrame(self, frame_index):
        prev_frame = self.cvGetGrayFrame(frame_index-1);
        this_frame = self.vbGetGrayFrame(frame_index);
        return cvDenseFlowFarneback(from_image=prev_frame, to_image=this_frame);

    def getFlowFramePolar(self, frame_index):
        """
        :param self:
        :param frame_index:
        :return: polar where polar[:,:,0] is amplitude, and polar[:,:,1] is angle
        """
        flow = self.getFlowFrame(frame_index);
        fx, fy = flow[:,:,0], flow[:,:,1];
        polar = np.zeros(size(flow));
        polar[:,:,0] = np.sqrt(fx * fx + fy * fy);
        polar[:,:,1] = np.arctan2(fy, fx) + np.pi
        return polar;

    def computeDirectogramPowers(self, bins=None, dead_zone= 0.05, density=None, save_if_computed=True, noise_floor_percentile=None, **kwargs):
        if(bins is None):
            bins = 128;
        if(noise_floor_percentile is None):
            noise_floor_percentile = 20;

        print(("Computing Flow Features with deadzone {}".format(dead_zone)))

        signal_dim = 128;
        m_histvals = np.zeros([signal_dim, self.n_frames(), 3]);

        flow_averages = np.zeros([self.n_frames(), 1]);
        sampling_rate=self.sampling_rate;
        duration = self.getDuration();
        nsamples = sampling_rate*duration;

        frame_start_times = np.linspace(0,duration,num=nsamples,endpoint=False);
        frame_index_floats = frame_start_times*self.sampling_rate;

        lastframe = self.cvGetGrayFrame(frame_index_floats[0]);

        start_timer=time.time();
        last_timer=start_timer;
        fcounter=0;
        counter = 0;

        for nf in range(len(frame_index_floats)):
            nextframe= self.cvGetGrayFrame(frame_index_floats[nf]);
            flow = cvDenseFlowFarneback(from_image=lastframe, to_image=nextframe);
            h, w = flow.shape[:2];
            fx, fy = flow[:,:,0], flow[:,:,1];

            # if(filter_median):
            #     fx = fx-np.median(fx.ravel());
            #     fy = fy-np.median(fy.ravel());
            #     assert(False), "SHOULDNT BE FILTERING MEDIAN! VESTIGIAL CODE"

            ang = np.arctan2(fy, fx) + np.pi
            amp = np.sqrt(fx*fx+fy*fy);

            winstarty = int(dead_zone*h);
            winendy = h-winstarty;
            winstartx = int(dead_zone*w);
            winendx = w-winstartx;
            angw = ang[winstarty:winendy, winstartx:winendx];
            ampw = amp[winstarty:winendy, winstartx:winendx];

            mask0 = (ampw>np.percentile(ampw,noise_floor_percentile)).astype(float);
            ahis0, cbinbounds = np.histogram(angw.ravel(), bins=bins, range=(0, 2 * np.pi),
                                             weights=mask0.ravel(), density=density);
            ahis1, cbinbounds1 = np.histogram(angw.ravel(), bins=bins, range=(0, 2 * np.pi),
                                             weights=ampw.ravel(), density=density);
            ahis2, cbinbounds2 = np.histogram(angw.ravel(), bins=bins, range=(0, 2 * np.pi),
                                             weights=np.power(ampw, 2).ravel(), density=density);

            m_histvals[:, counter, 0] = ahis0;
            m_histvals[:, counter, 1] = ahis1;
            m_histvals[:, counter, 2] = ahis2;

            lastframe=nextframe;
            counter+=1;
            fcounter+=1;

            if(not (fcounter%50)):
                if((time.time()-last_timer)>10):
                    last_timer=time.time();
                    print(("{}%% done after {} seconds...".format(100.0*truediv(fcounter,len(frame_index_floats)), last_timer-start_timer)));
        params = dict( bins = bins,
                        deadzone=dead_zone,
                        density=density);
        params.update(kwargs);
        self.setFeature(name='directogram_powers', value=m_histvals, params=params);
        if(save_if_computed):
            self.save(features_to_save=['directogram_powers']);
        return m_histvals;

    ######################### Other Features #############################

    def getDirectogramPowers(self, force_recompute=False, **kwargs):
        feature_name = 'directogram_powers';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            flow_powers = self.computeDirectogramPowers(**kwargs);
        return self.getFeature(feature_name);

    # def getDirectogram(self, bins = None, weights=None, density=None, force_recompute=False, save_if_computed=True, **kwargs):
    def getDirectogram(self, **kwargs):
        feature_name = 'directogram';
        force_recompute = kwargs.get('force_recompute');
        if((not self.hasFeature(feature_name)) or force_recompute):
            flow_powers = self.getFeature('directogram_powers');
            fh = flow_powers[:,:,1];
            self.setFeature(name='directogram', value=fh, params=kwargs);
        return self.getFeature(feature_name);

    def getVisualTempo(self, force_recompute=None, **kwargs):
        feature_name = 'visual_tempo';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            vbe = self.getFeature('local_rhythmic_saliency');
            result = librosa.beat.estimate_tempo(onset_envelope=vbe, sr=self.sampling_rate, hop_length=1, **kwargs);
            self.setFeature(name=feature_name, value=result, params=kwargs);
        return self.getFeature(feature_name);

    def getVisualTempogram(self, window_length=None, force_recompute=None, norm_columns = None, **kwargs):
        """

        :param self:
        :param window_length: in seconds
        :param force_recompute:
        :param kwargs:
        :return:
        """
        feature_name = 'visual_tempogram';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            if(window_length is None):
                window_length = DEFAULT_TEMPOGRAM_WINDOW_SECONDS;
            params = kwargs;
            params.update({'force_recompute': force_recompute});
            vbe = self.computeImpactEnvelope(cut_suppression_seconds = None);
            onset_envelope = vbe;
            win_length = int(round(window_length * self.sampling_rate));
            sr = self.sampling_rate;
            hop_length = 1;

            center=kwargs.get('center');
            if(center is None):
                center=True;
            window='hann'
            norm=np.inf;
            ac_window = librosa.filters.get_window(window, win_length, fftbins=True)

            # Center the autocorrelation windows
            n = len(onset_envelope)

            if center:
                onset_envelope = np.pad(onset_envelope, int(win_length // 2),
                                        mode='linear_ramp', end_values=[0, 0])
            # Carve onset envelope into frames
            odf_frame = librosa.util.frame(onset_envelope,
                                   frame_length=win_length,
                                   hop_length=hop_length)
            # Truncate to the length of the original signal
            if center:
                odf_frame = odf_frame[:, :n]

            odf_frame = librosa.util.normalize(odf_frame,axis=0,norm=norm)
            if(norm_columns is None):
                norm_columns = True;

            if(norm_columns):
                # Window, autocorrelate, and normalize
                result = librosa.util.normalize(
                    librosa.core.audio.autocorrelate(odf_frame * ac_window[:, np.newaxis], axis=0), norm=norm, axis=0);
            else:
                result = librosa.core.audio.autocorrelate(odf_frame * ac_window[:, np.newaxis], axis=0);
                result = np.true_divide(result, np.max(result.ravel()));

            tempo_bpms = librosa.tempo_frequencies(result.shape[0], hop_length=hop_length, sr=sr)
            self.setFeature(name='tempogram_bpms', value=tempo_bpms);
            ###########
            self.setFeature(name=feature_name, value=result, params=params);

        return self.getFeature(feature_name);

    def getVisibleImpactEnvelope(self, force_recompute=False, **kwargs):
        feature_name = 'impact_envelope';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            result = self.computeImpactEnvelope(forward=True, backward = False, **kwargs);
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    def getForwardVisibleImpactEnvelope(self, force_recompute=False, **kwargs):
        """
        Same as getVisibleImpactEnvelope by default.
        :param self:
        :param force_recompute:
        :param kwargs:
        :return:
        """
        feature_name = 'forward_visual_impact_envelope';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            result = self.computeImpactEnvelope(forward=True, backward = False, **kwargs);
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    def getBackwardVisibleImpactEnvelope(self, force_recompute=False, **kwargs):
        feature_name = 'backward_visual_impact_envelope';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            result = self.computeImpactEnvelope(forward=False, backward = True, **kwargs);
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    def getBothWayVisibleImpactEnvelope(self, force_recompute=False, **kwargs):
        feature_name = 'both_way_visual_impact_envelope';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            result = self.computeImpactEnvelope(forward=True, backward = True, **kwargs);
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    def getVisibleImpactEnvelopePowers(self, force_recompute=False, **kwargs):
        feature_name = 'impact_envelope_powers';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            result0 = self.computeImpactEnvelope(power= 0, **params);
            result = np.zeros([len(result0), 3]);
            result[:, 0] = result0;
            result[:, 1] = self.computeImpactEnvelope(power= 1, **params);
            result[:, 2] = self.computeImpactEnvelope(power=2, **params);
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    def getCutTimes(self):
        return Event.ToStartTimes(self.getCutEvents());

    def getCutEvents(self, force_recompute=False, **kwargs):
        """
        Hacky estimate of cuts in a video
        :param self:
        :param force_recompute:
        :param kwargs:
        :return:
        """
        feature_name = 'cut_events';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            powers = self.getFeature('directogram_powers');
            im0 = powers[:, :, 0];
            im2 = powers[:, :, 2];
            imc = np.true_divide(im2, 0.05 + im0);

            medsig = np.median(imc, 0);
            medall = np.median(imc);
            cutsig = np.true_divide(medsig, medall)
            cut_detection_ratio = kwargs.get('cut_detection_ratio');
            if(cut_detection_ratio is None):
                cut_detection_ratio=CUT_DETECTION_RATIO;
            clipsig = (cutsig > cut_detection_ratio).astype(float);
            clip_floorsig = (cutsig > CUT_DETECTION_FLOOR).astype(float);
            clipsig = np.multiply(clipsig, clip_floorsig);

            einds = np.flatnonzero(clipsig);
            etimes = einds * truediv(1.0, self.sampling_rate);
            ev = Event.FromStartTimes(etimes, type='cut');
            ev = self.visualBeatsFromEvents(ev);
            evout = []
            for e in range(len(ev)):
                ev[e].setInfo('frame', einds[e]);
            result = ev;
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    def visualBeatsFromEvents(self, events):
        def downsample_hist(sig, levels):
            nperbin = np.power(2, levels);
            rshp = sig.reshape(-1, nperbin);
            return rshp.sum(axis=1);
        if(self.hasFeature('impact_envelope')):
            svbe=self.getFeature('impact_envelope');
        else:
            svbe=self.computeImpactEnvelope(cut_suppression_seconds = None);
        flow_powers = self.getFeature('directogram_powers');
        vis_tempogram = self.getFeature('visual_tempogram');

        vbeats = [];
        for e in events:
            b = VisualBeat.FromEvent(e);
            ei = int(round(b.start*self.sampling_rate*VB_UPSAMPLE_FACTOR));
            b.weight = svbe[ei];

            histsize = 128/int(np.power(2,HISTOGRAM_DOWNSAMPLE_LEVELS))
            histslice = np.zeros([histsize, HISTOGRAM_FRAMES_PER_BEAT]);
            histslice = np.squeeze(np.mean(histslice, 1));
            b.flow_histogram = downsample_hist(sig=histslice, levels=HISTOGRAM_DOWNSAMPLE_LEVELS);
            b.flow_histogram = np.divide(b.flow_histogram, np.sum(b.flow_histogram));
            b.local_autocor = vis_tempogram[:,ei];
            b.local_autocor = b.local_autocor/np.max(b.local_autocor);
            b.sampling_rate=self.sampling_rate;
            vbeats.append(b);
        return vbeats;

    def getVisualBeatTimes(self, **kwargs):
        return Event.ToStartTimes(self.getVisualBeats(**kwargs));

    def getDirectionalFlux(self,
                            f_sigma=None,
                            median_kernel=None,
                            power=None,
                            **kwargs):
        """
        The visual impact complement of a spectral flux matrix
        :param self:
        :param f_sigma: sigma for the gaussian used to filter, using sp.ndimage.filters.gaussian_filter
        :param median_kernel: used in sp.signal.medfilt
        :param power: Which power of the flow to use. Usually use 1.
        :param kwargs:
        :return:
        """

        def d_x(im):
            d_im = np.zeros(im.shape);
            d_im[:, 0] = im[:, 0];
            for c in range(1, im.shape[1]):
                d_im[:, c] = im[:, c] - im[:, c - 1];
            return d_im;

        if (f_sigma is None):
            f_sigma = [5, 3];
        if (median_kernel is None):
            median_kernel = [3, 3];

        if(power is None):
            power = 1;

        powers = self.getFeature('directogram_powers');
        im = powers[:,:,power].copy();
        if (f_sigma is not None):
            im = sp.ndimage.filters.gaussian_filter(input=im, sigma=f_sigma, order=0);

        im = sp.signal.medfilt(im, median_kernel);
        return d_x(im);

    def computeImpactEnvelope(self,
                             forward=True,
                             backward = False,
                             f_sigma=None,
                             median_kernel=None,
                             highpass_window_seconds= 0.8,
                             cut_percentile=99,
                             power=None,
                             crop = None,
                             normalize = True,
                             **kwargs):
        """

        :param self:
        :param forward: include impact going forward in time
        :param backward: include impact going backward in time
        :param f_sigma: sigma for the gaussian used to filter, using sp.ndimage.filters.gaussian_filter
        :param median_kernel: used in sp.signal.medfilt
        :param highpass_window_seconds: highpass window size in seconds
        :param cut_percentile: percentile above which to consider cuts and to clip
        :param power: Which power of the flow to use. Usually use 1.
        :param crop:
        :param kwargs:
        :return:
        """

        upsample_factor = kwargs.get('upsample_factor');
        inputargs = dict(f_sigma=f_sigma,
                         median_kernel=median_kernel,
                         power=power,
                         crop = crop);

        inputargs.update(kwargs);
        im_d = self.getDirectionalFlux(**inputargs);

        if(forward and backward):
            im = np.fabs(im_d);
        elif(forward):
            im = -im_d;
            im = np.clip(im, 0, None)
        elif(backward):
            im = im_d;
            im = np.clip(im, 0, None)
        else:
            assert(False), "Must be at least one of either forward or backward."

        vimpact = np.squeeze(np.mean(im, 0));
        sampling_rate = self.sampling_rate;
        if(upsample_factor is not None and (upsample_factor>1)):
            newlen = upsample_factor * len(vimpact);
            sampling_rate = upsample_factor*sampling_rate;
            vimpact = sp.signal.resample(vimpact, newlen);

        if(highpass_window_seconds):
            order = kwargs.get('highpass_order');
            if(order is None):
                order = 5;
            cutoff = truediv(1.0, highpass_window_seconds);
            normal_cutoff = cutoff / (sampling_rate*0.5);
            b, a = sp.signal.butter(order, normal_cutoff, btype='high', analog=False)
            vimpact = sp.signal.filtfilt(b, a, vimpact);


        normfactor = np.max(np.fabs(vimpact[:]));

        if (cut_percentile is not None):
            fx = np.fabs(vimpact);
            pv = np.percentile(fx, cut_percentile);
            pvlow = np.percentile(fx, cut_percentile-1);
            normfactor = pv;
            ptile = (vimpact > pv).astype(float);
            pntile = (vimpact < -pv).astype(float);
            pboth = ptile+pntile;
            einds = np.flatnonzero(pboth);
            lastind = -2;
            for j in range(len(einds)):
                if(einds[j]==(lastind+1)):
                    vimpact[einds[j]]=0;
                else:
                    vimpact[einds[j]]=pvlow;

        if(normalize):
            vimpact = np.true_divide(vimpact, normfactor);
        return vimpact;

    # 0.8, cut_suppression_seconds = 0.4,
    def computeImpactEnvelopeOld(self, f_sigma=None, median_kernel=None, highpass_window_seconds= 0.8, cut_percentile=99, power=None, crop = None, normalize=None, **kwargs):
        """
        I believe this is the version of the function that I used in the original paper. Keeping it around for record.
        :param self:
        :param f_sigma:
        :param median_kernel:
        :param highpass_window_seconds:
        :param cut_percentile:
        :param power:
        :param crop:
        :param normalize:
        :param kwargs:
        :return:
        """
        def d_x(im):
            # d_im = np.zeros(im.shape, dtype=np.float128);
            d_im = np.zeros(im.shape);
            d_im[:, 0] = im[:, 0];
            for c in range(1, im.shape[1]):
                d_im[:, c] = im[:, c] - im[:, c - 1];
            return d_im;

        if (f_sigma is None):
            f_sigma = [5, 3];
        if (median_kernel is None):
            median_kernel = [3, 3];

        if(power is None):
            power = 1;

        upsample_factor = kwargs.get('upsample_factor');

        powers = self.getFeature('directogram_powers');
        print(("computing impact env with power {}".format(power)));
        im = powers[:,:,power].copy();

        if (f_sigma is not None):
            im = sp.ndimage.filters.gaussian_filter(input=im, sigma=f_sigma, order=0);

        im = sp.signal.medfilt(im, median_kernel);
        im = -d_x(im);
        im = np.clip(im, 0, None)

        svbe = np.squeeze(np.mean(im, 0));
        sampling_rate = self.sampling_rate;

        if(upsample_factor is not None and (upsample_factor>1)):
            newlen = upsample_factor * len(svbe);
            sampling_rate = upsample_factor*sampling_rate;
            svbe = sp.signal.resample(svbe, newlen);

        if(highpass_window_seconds):
            order = kwargs.get('highpass_order');
            if(order is None):
                order = 5;
            cutoff = truediv(1.0, highpass_window_seconds);
            normal_cutoff = cutoff / (sampling_rate*0.5);
            b, a = sp.signal.butter(order, normal_cutoff, btype='high', analog=False)
            svbe = sp.signal.filtfilt(b, a, svbe);


        normfactor = np.max(np.fabs(svbe[:]));

        if (cut_percentile is not None):
            fx = np.fabs(svbe);
            pv = np.percentile(fx, cut_percentile);
            pvlow = np.percentile(fx, cut_percentile-1);
            normfactor = pv;
            ptile = (svbe > pv).astype(float);
            pntile = (svbe < -pv).astype(float);
            pboth = ptile+pntile;
            einds = np.flatnonzero(pboth);
            lastind = -2;
            for j in range(len(einds)):
                if(einds[j]==(lastind+1)):
                    svbe[einds[j]]=0;
                else:
                    svbe[einds[j]]=pvlow;

        if(normalize is not False):
            svbe = np.true_divide(svbe, normfactor);
        return svbe;

    def getVisibleImpacts(self, force_recompute=False, include_cut_events = None, **kwargs):
        feature_name = 'visible_impacts';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            svbe = self.getFeature('impact_envelope', **kwargs);
            upsample_factor = kwargs.get('upsample_factor');
            if(upsample_factor is None):
                upsample_factor = 1;
            u_sampling_rate = self.sampling_rate*upsample_factor;

            peak_params = self._getDefaultPeakPickingTimeParams();
            peak_params.update(kwargs); # if params given in arguments, those will override the defaults here.

            v_events = Event.FromSignalPeaks(signal=svbe, sampling_rate=u_sampling_rate, **peak_params);
            if(include_cut_events):
                cut_events = self.getFeature('cut_events');
                v_events = v_events + cut_events;
                Event.Sort(v_events);
            result = self.visualBeatsFromEvents(v_events);

            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    def getForwardVisibleImpacts(self, force_recompute=False, include_cut_events = None, **kwargs):
        feature_name = 'forward_visual_beats';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            local_saliency = self.getFeature('forward_visual_impact_envelope', **kwargs);
            upsample_factor = kwargs.get('upsample_factor');
            if(upsample_factor is None):
                upsample_factor = 1;
            u_sampling_rate = self.sampling_rate*upsample_factor;
            v_events = Event.FromSignalPeaks(signal=local_saliency, sampling_rate=u_sampling_rate, event_type= 'forward', **kwargs);
            if(include_cut_events):
                cut_events = self.getFeature('cut_events');
                v_events = v_events + cut_events;
                Event.Sort(v_events);
            result = Event.SetDirections(v_events, direction=Event.DIRECTION_FORWARD);
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);

    def getBackwardVisibleImpacts(self, force_recompute=False, include_cut_events = None, **kwargs):
        feature_name = 'backward_visual_beats';
        if ((not self.hasFeature(feature_name)) or force_recompute):
            params = kwargs;
            local_saliency = self.getFeature('backward_visual_impact_envelope', **kwargs);
            upsample_factor = kwargs.get('upsample_factor');
            if(upsample_factor is None):
                upsample_factor = 1;
            u_sampling_rate = self.sampling_rate*upsample_factor;
            v_events = Event.FromSignalPeaks(signal=local_saliency, sampling_rate=u_sampling_rate, **kwargs);
            if(include_cut_events):
                cut_events = self.getFeature('cut_events');
                v_events = v_events + cut_events;
                Event.Sort(v_events);
            # result = self.visualBeatsFromEvents(v_events);
            result = Event.SetDirections(v_events, direction=Event.DIRECTION_BACKWARD);
            self.setFeature(name=feature_name, value=result, params=params);
        return self.getFeature(feature_name);


    def findAccidentalDanceSequences(self, target_n_beats = 7, n_samples=25, delta_range = None):
        if(delta_range is None):
            delta_range = [0.02, 0.5];

        deltas = np.linspace(delta_range[0], delta_range[1], num=n_samples, endpoint=True);
        deltapick = delta_range[0];
        sequences = [];
        for i, delta in enumerate(deltas):
            peak_vars = self._getDefaultPeakPickingTimeParams(delta=delta);
            sequences = self.getVisualBeatSequences(peak_vars=peak_vars, print_summary=False);
            # print("Delta {} has top sequence with {} beats".format(delta, len(sequences[0])));
            if(len(sequences[0])<=target_n_beats):
                deltapick = delta;
                break;
        print(("Selected delta value {}".format(deltapick)));
        return sequences;


    def getVisualBeatSequences(self,
                               search_tempo=None,
                               target_period=None,
                               search_window=0.75,
                               min_beat_limit=None,
                               max_beat_limit=None,
                               unary_weight=None,
                               binary_weight=None,
                               break_on_cuts=None,
                               peak_vars=None,
                               time_range=None,
                               n_return = None,
                               unsorted = False,
                               print_summary = True,
                               **kwargs):
        """

        :param self:
        :param search_tempo: optional tempo you would like to find visual beats at
        :param target_period: optional target period you would like to use for finding beats. Ignored if search tempo is provided.
        :param search_window: longest amount of time (seconds) allowed between beats before a segment is broken into multiple segments
        :param min_beat_limit: only consider sequences with
        :param unary_weight:
        :param binary_weight:
        :param break_on_cuts:
        :param peak_vars:
        :param kwargs:
        :return:
        """
        if (peak_vars is not None):
            # impacts = self.getFeature('visible_impacts', force_recompute=True, **peak_vars);
            impacts = self.getVisualBeats(force_recompute = True, **peak_vars);
        else:
            # impacts = self.getFeature('visible_impacts');
            impacts = self.getVisualBeats();

        if(time_range is not None):
            impactseg = [];
            for i in impacts:
                if(i.start>time_range[0] and i.start < time_range[1]):
                    impactseg.append(i);
            impacts = impactseg;


        if (search_tempo is not None):
            tempo = search_tempo;
            beat_time = np.true_divide(60.0, tempo);
            sequences = VisualBeat.PullOptimalPaths_Basic(impacts, target_period=beat_time, unary_weight=unary_weight,
                                                      binary_weight=binary_weight, break_on_cuts=break_on_cuts,
                                                      window_size=search_window);

        elif(target_period is not None):
            sequences = VisualBeat.PullOptimalPaths_Basic(impacts, target_period=target_period, unary_weight=unary_weight,
                                                          binary_weight=binary_weight, break_on_cuts=break_on_cuts,
                                                          window_size=search_window);
        else:
            sequences = VisualBeat.PullOptimalPaths_Autocor(impacts, unary_weight=unary_weight, binary_weight=binary_weight,
                                                        break_on_cuts=break_on_cuts, window_size=search_window);

        r_sequences = [];

        if(min_beat_limit is None):
            min_beat_limit = 2;
        if(max_beat_limit is None):
            max_beat_limit = len(impacts)+1;

        for S in sequences:
            if ((len(S) > min_beat_limit) and (len(S) < max_beat_limit)):
                r_sequences.append(S);

        if(not unsorted):
            r_sequences.sort(key=len, reverse=True);
            if(n_return is not None):
                r_sequences = r_sequences[:n_return];
        if(print_summary):
            print(("{} segments".format(len(r_sequences))));
            for s in range(len(r_sequences)):
                print(("Segment {} has {} beats".format(s, len(r_sequences[s]))));

        return r_sequences;


    def printVisualBeatSequences(self,
                                 search_tempo=None,
                                 target_period=None,
                                 search_window=None,
                                 min_beat_limit=None,
                                 max_beat_limit=None,
                                 unary_weight=None,
                                 binary_weight=None,
                                 break_on_cuts=None,
                                 peak_vars=None,
                                 n_return = None,
                                 time_range=None, **kwargs):
        """

        :param self:
        :param target_period:
        :param search_tempo:
        :param search_window:
        :param beat_limit:
        :param unary_weight:
        :param binary_weight:
        :param break_on_cuts:
        :param n_return:
        :param peak_vars:
        :param time_range:
        :param kwargs:
        :return:
        """

        sorted = True;

        sequence_args = dict(
            search_tempo=search_tempo,
            target_period=target_period,
            search_window=search_window,
            min_beat_limit=min_beat_limit,
            max_beat_limit=max_beat_limit,
            unary_weight=unary_weight,
            binary_weight=binary_weight,
            break_on_cuts=break_on_cuts,
            peak_vars=peak_vars,
            n_return=n_return,
            time_range=time_range);

        seqs = self.getVisualBeatSequences(**sequence_args)
        print(("sequence arguments were:\n{}".format(sequence_args)));
        print(("There were {} sequences total".format(len(seqs))));
        nclips = 0;
        rsegments = [];
        for S in seqs:
            if (len(S) > 1):
                nclips = nclips + 1;
                rsegments.append(S);

        # rsegments.sort(key=len, reverse=True);
        # if (n_return is not None):
        #     rsegments = rsegments[:n_return];

        Event.PlotSignalAndEvents(self.getFeature('impact_envelope'),  sampling_rate=self.sampling_rate*VB_UPSAMPLE_FACTOR, events=rsegments[0],  time_range=time_range);
        return rsegments;

    def plotEvents(self, events, time_range = 'default', **kwargs):
        time_range_use = time_range;
        if(time_range.lower() == 'default'):
            time_range_use = [0,0];
            time_range_use[0] = events[0].start-1;
            time_range_use[1] = events[-1].start+1;

        signal = self.getFeature('local_rhythmic_saliency');
        mplt = Event.PlotSignalAndEvents(signal, sampling_rate=self.sampling_rate * VB_UPSAMPLE_FACTOR, events=events, time_range=time_range_use, **kwargs);
        plt.xlabel('Time (s)')
        return mplt;

    def plotCutEvents(self, **kwargs):
        signal = self.getFeature('impact_envelope');
        events = self.getFeature('cut_events', **kwargs);
        Event.PlotSignalAndEvents(signal, sampling_rate=self.sampling_rate*VB_UPSAMPLE_FACTOR, events=events, **kwargs);

    def plotVisibleImpacts(self, **kwargs):
        signal = self.getFeature('impact_envelope');
        events = self.getFeature('visible_impacts', **kwargs);
        Event.PlotSignalAndEvents(signal, sampling_rate=self.sampling_rate*VB_UPSAMPLE_FACTOR, events=events, **kwargs);
        plt.title('Impact Envelope & Visual Beats')
        plt.xlabel('Time (s)')
        plt.ylabel('Impact Strength')

    def plotImpactEnvelope(self, **kwargs):
        signal = self.getFeature('local_rhythmic_saliency');
        # events = self.getFeature('visual_beats', **kwargs);
        events = None;
        Event.PlotSignalAndEvents(signal, sampling_rate=self.sampling_rate*VB_UPSAMPLE_FACTOR, events=events, **kwargs);
        plt.title('Impact Envelope & Visual Beats')
        plt.xlabel('Time (s)')
        plt.ylabel('Impact Strength')

    def plotVisualBeats(self, **kwargs):
        signal = self.getFeature('local_rhythmic_saliency');
        events = self.getFeature('visual_beats', **kwargs);
        Event.PlotSignalAndEvents(signal, sampling_rate=self.sampling_rate*VB_UPSAMPLE_FACTOR, events=events, **kwargs);
        plt.title('Impact Envelope & Visual Beats')
        plt.xlabel('Time (s)')
        plt.ylabel('Impact Strength')


    def loadFlowFeatures(self):
        self.load(features_to_load=['directogram_powers', 'directogram']);

    FEATURE_FUNCS['local_rhythmic_saliency'] = getLocalRhythmicSaliency;
    FEATURE_FUNCS['directogram_powers'] = getDirectogramPowers;
    FEATURE_FUNCS['directogram'] = getDirectogram;
    FEATURE_FUNCS['impact_envelope'] = getVisibleImpactEnvelope;
    FEATURE_FUNCS['impact_envelope_powers'] = getVisibleImpactEnvelopePowers;
    FEATURE_FUNCS['visible_impacts'] = getVisibleImpacts;
    FEATURE_FUNCS['visual_beats'] = getVisualBeats;
    # FEATURE_FUNCS['backward_visual_beats'] = getBackwardVisualBeats;
    # FEATURE_FUNCS['forward_visual_beats'] = getForwardVisualBeats;
    FEATURE_FUNCS['backward_visual_impact_envelope'] = getBackwardVisibleImpactEnvelope;
    FEATURE_FUNCS['both_way_visual_impact_envelope'] = getBothWayVisibleImpactEnvelope;
    FEATURE_FUNCS['forward_visual_impact_envelope'] = getForwardVisibleImpactEnvelope;
    FEATURE_FUNCS['directional_flux'] = getDirectionalFlux;
    FEATURE_FUNCS['visual_tempogram'] = getVisualTempogram;
    FEATURE_FUNCS['cut_events']=getCutEvents;





from .Event import *

class VisualBeat(Event):
    def VBOBJECT_TYPE(self):
        return 'VisualBeat';

    def __init__(self, start=None, type=None, weight=None, index=None, unrolled_start = None, direction=0):
        Event.__init__(self, start=start, type=type, weight=weight, index=index, unrolled_start=unrolled_start, direction=direction);

    def initializeBlank(self):
        Event.initializeBlank(self);
        self.flow_histogram = None;
        self.local_autocor = None;
        self.sampling_rate = None;

    def __str__(self):
        return "start:{}\ntype:{}\nweight:{}\nindex:{}\nunrolled_start:{}\nis_active:{}\n".format(self.start, self.type, self.weight, self.index, self.unrolled_start, self.is_active);

    # def toGUIDict(self, is_active=1):
    #     return dict(start=self.start, index=self.index, is_active=None)

    def toDictionary(self):
        d=Event.toDictionary(self);
        # d['start']=self.start;
        # d['type']=self.type;
        # d['weight']=self.weight;
        # d['index']=self.index;
        # d['unrolled_start']=self.unrolled_start;
        d['flow_histogram']=self.flow_histogram;
        d['local_autocor']=self.local_autocor;
        d['sampling_rate']=self.sampling_rate;
        return d;

    def initFromDictionary(self, d):
        Event.initFromDictionary(self, d);
        self.start = d['start'];
        self.type = d['type'];
        self.weight = d['weight'];
        self.index = d['index'];
        self.unrolled_start = d['unrolled_start'];
        self.flow_histogram = d.get('flow_histogram');
        self.local_autocor = d.get('local_autocor');
        self.sampling_rate = d.get('sampling_rate');


    def clone(self, start=None):
        if(start):
            newv = VisualBeat(start = start, type=self.type, weight=self.weight, index=self.index);
        else:
            newv = VisualBeat(start = self.start, type=self.type, weight=self.weight, index=self.index);
        newv.flow_histogram = self.flow_histogram.copy();
        newv.local_autocor = self.local_autocor.copy();
        newv.sampling_rate = self.sampling_rate;
        return newv;

    @staticmethod
    def FromEvent(e):
        if(isinstance(e,VisualBeat)):
            return e.clone();
        else:
            return VisualBeat(start=e.start, type=e.type, weight=e.weight, index=e.index);

    @staticmethod
    def time_window_func(max_separation, break_on_cuts=None):
        def window_func(a, b):
            if(break_on_cuts and (a.type=='cut' or b.type=='cut')):
                return False;
            if(np.fabs(a.start-b.start)<max_separation):
                return True;
            else:
                return False;
        return window_func;

    @staticmethod
    def tempo_binary_objective(target_period, binary_weight=None):
        if(binary_weight is None):
            binary_weight = 1.0;
        def objective_func(a, b):
            T = np.fabs(a.start-b.start);
            return -np.power(np.log(truediv(T,target_period)), 2.0)*binary_weight;
        return objective_func;

    @staticmethod
    def autocor_binary_objective(binary_weight=None, **kwargs):
        if (binary_weight is None):
            binary_weight = 1.0;

        def objective_func(a, b):
            T = np.fabs(a.start - b.start);

            # assert(a.sampling_rate), "b has no sampling rate"
            # assert(b.sampling_rate), "a has no sampling rate"
            abin = int(round(np.true_divide(T,a.sampling_rate)));
            score = (a.local_autocor[abin] - 1);
            if (T < 0.25):
                score = -1;
            if(T>3.75):
                score=-1;
            return binary_weight*score;

        return objective_func;

    @staticmethod
    def angle_binary_objective(binary_weight=None, absolute=None):
        if (binary_weight is None):
            binary_weight = 1.0;
        # if (angle_weight is None):
        #     angle_weight = 0.0;

        def objective_func(a, b):
            # T = np.fabs(a.start - b.start);
            # tempo_score = -np.power(np.log(truediv(T, target_period)), 2.0) * binary_weight;
            # should mask out unary contribution from orthogonal angles
            if(absolute):
                return binary_weight * (np.fabs(np.dot(a.flow_histogram, b.flow_histogram)) - 0.70710678118);  # cos 45 degrees
            else:
                return binary_weight * (np.dot(a.flow_histogram, b.flow_histogram));  # cos 45 degrees
        return objective_func;

    @staticmethod
    def Double(events, type=None):
        doubled = [];
        for e in range(1, len(events)):
            halfstart = 0.5 * (events[e].start + events[e - 1].start);
            newhevent = events[e].clone(start=halfstart);
            if(type is not None):
                newhevent.type = type;
            doubled.append(newhevent);
            doubled.append(events[e]);
        return doubled;

    @staticmethod
    def weight_unary_objective(unary_weight=None):
        if(unary_weight is None):
            unary_weight = 1.0;
        def getweight_func(b):
            return unary_weight*b.weight;
        return getweight_func;

    @staticmethod
    def PullOptimalPaths_Basic(vis_beats, target_period, unary_weight=None, binary_weight=None, window_size=None, break_on_cuts = None):
        if(window_size is None):
            window_size = DEFAULT_WINDOW_FACTOR*target_period;
        binary_objective = VisualBeat.tempo_binary_objective(target_period=target_period, binary_weight = binary_weight);
        unary_objective = VisualBeat.weight_unary_objective(unary_weight=unary_weight);
        window_function = VisualBeat.time_window_func(max_separation = window_size, break_on_cuts=break_on_cuts);
        return VisualBeat.DynamicProgramOptimalPaths(vis_beats=vis_beats,
                                                     unary_objective_func=unary_objective,
                                                     binary_objective_func=binary_objective,
                                                     window_func=window_function);

    @staticmethod
    def PullOptimalPaths(vis_beats, unary_fn=None, binary_fn=None, window_fn=None,  target_period=None, unary_weight=None, binary_weight=None, window_size=None,
                               break_on_cuts=None):
        if (window_size is None):
            window_size = DEFAULT_WINDOW_FACTOR * target_period;

        if(binary_fn == 'autocor'):
            binary_objective = VisualBeat.autocor_binary_objective(binary_weight=binary_weight);
        elif(binary_fn == 'angle'):
            binary_objective = VisualBeat.angle_binary_objective(binary_weight=binary_weight);
        else:
            binary_objective = VisualBeat.tempo_binary_objective(target_period=target_period, binary_weight=binary_weight);

        unary_objective = VisualBeat.weight_unary_objective(unary_weight=unary_weight);
        window_function = VisualBeat.time_window_func(max_separation=window_size, break_on_cuts=break_on_cuts);
        return VisualBeat.DynamicProgramOptimalPaths(vis_beats=vis_beats,
                                                     unary_objective_func=unary_objective,
                                                     binary_objective_func=binary_objective,
                                                     window_func=window_function);

    @staticmethod
    def PullOptimalPaths_Autocor(vis_beats, unary_weight=None, binary_weight=None, window_size=None,
                               break_on_cuts=None, **kwargs):
        if (window_size is None):
            # assert(False), 'no window size provided'
            # window_size = DEFAULT_WINDOW_FACTOR * target_period;
            window_size = 200;
            AWARN('NO WINDOWSIZE PROVIDED! PullOptimalPaths_Autocor');
        binary_objective = VisualBeat.autocor_binary_objective(binary_weight=binary_weight);
        unary_objective = VisualBeat.weight_unary_objective(unary_weight=unary_weight);
        window_function = VisualBeat.time_window_func(max_separation=window_size, break_on_cuts=break_on_cuts);
        return VisualBeat.DynamicProgramOptimalPaths(vis_beats=vis_beats,
                                                     unary_objective_func=unary_objective,
                                                     binary_objective_func=binary_objective,
                                                     window_func=window_function);

    @staticmethod
    def DynamicProgramOptimalPaths(vis_beats, unary_objective_func, binary_objective_func, window_func):
        class Node(object):
            def __init__(self, object, prev_node=None):
                self.object = object;
                self.prev_node = prev_node;
                self.cum_score=None;

        nodes = [];
        beats = Event.GetSorted(vis_beats);
        Event.ApplyIndices(beats);

        for b in beats:
            nodes.append(Node(object=b));

        nodes[0].prev_node = None;
        nodes[0].cum_score = unary_objective_func(nodes[0].object);
        current_segment = [];
        segments = [];
        for n in range(1,len(nodes)):
            current_node = nodes[n];
            current_segment.append(current_node);
            options = [];
            j = n-1;
            while(j>=0 and window_func(current_node.object,nodes[j].object)):
                options.append(nodes[j]);
                j = j-1;
            if(len(options)==0):
                current_node.prev_node=None;
                current_node.cum_score=unary_objective_func(current_node.object);
                segments.append(current_segment);
                current_segment = [];
            else:
                best_choice = options[0];
                best_score = options[0].cum_score+binary_objective_func(current_node.object, best_choice.object);
                for o in range(1,len(options)):
                    score = options[o].cum_score+binary_objective_func(current_node.object, options[o].object);
                    if(score>best_score):
                        best_choice=options[o];
                        best_score=score;
                current_node.prev_node = best_choice;
                current_node.cum_score = best_score+unary_objective_func(current_node.object);
        if(len(current_segment)>0):
            segments.append(current_segment);
        sequences = [];
        for S in segments:
            seq = [];
            max_node = S[0];
            max_score = max_node.cum_score;
            for n in range(len(S)):
                if(S[n].cum_score>max_score):
                    max_node = S[n];
                    max_score = max_node.cum_score;

            trace_node = max_node;
            while(trace_node.prev_node is not None):
                seq.append(trace_node.object);
                trace_node=trace_node.prev_node;
            seq.reverse();
            sequences.append(seq);

        return sequences;